################################################################################
# Licensed Materials - Property of IBM Copyright IBM Corporation 2012. All Rights Reserved.
# U.S. Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP
# Schedule Contract with IBM Corp.
################################################################################

require "devops_build/providers/build_system_provider"
require "devops_build/providers/mock_rtc_provider"
require "ant"
require "rake"
include Rake::DSL
require "java"
require "devops_build/libs/sccd-ant-rtc-tasks-1.0.0.jar"
require "devops_build/libs/iwdapi-2.0.5.v20121019.jar"
require "devops_build/libs/jetty-client-7.5.4.v20111024.jar"
require "devops_build/libs/jetty-http-7.5.4.v20111024.jar"
require "devops_build/libs/jetty-io-7.5.4.v20111024.jar"
require "devops_build/libs/jetty-util-7.5.4.v20111024.jar"
require "devops_build/libs/JSON4J.jar"
require "devops_build/libs/sccd-ant-rtc-cloudlogs-tasks-1.0.0.jar"
require "devops_build/libs/slf4j-api-1.7.2.jar"
require "devops_build/libs/slf4j-log4j12-1.7.2.jar"


Dir["/opt/ibm/rtc_build/buildtoolkit/*.jar"].each { |jar| require jar }

module DevopsBuild
  module Providers
    ################################################################################
    # Globals                                                                      #
    ################################################################################

    ################################################################################
    # Class Definition                                                             #
    ################################################################################

    # RtcBuildSystemProvider - an implementation of the BuildSystemProvider
    # interface to interact with the Rational Team Concert build system.
    #                with the HSLT Cloud.
    #
    # For documentation on what each method does, please see the comments in 
    # com.ibm.devops.cloud/lib/build/providers/build_system_provider.rb
    #
    class RtcBuildSystemProvider < BuildSystemProvider
  
  
      ################################################################################
      # Class Methods
      ################################################################################
    
      # Returns an instance of RtcBuildSystemProvider or MockRtcBuildSystemProvider
      # depending on if the first argument (use_mock) is true.  For local builds this
      # lets you use a mock provider that pretends to interact with RTC.
      #  
      # Params
      # ======
      #   * use_mock - (Boolean) If true, return a mock provider
      # 
      def RtcBuildSystemProvider.newFor(use_mock, repository_url, user_id, password_hash={})
        if use_mock
          return DevopsBuild::Providers::MockRtcBuildSystemProvider.new(repository_url, user_id, password_hash)
        else
          return DevopsBuild::Providers::RtcBuildSystemProvider.new(repository_url, user_id, password_hash)
        end
      end

      ################################################################################
      # Instance Methods
      ################################################################################
      
      # Constructs a new RtcBuildSystemProvider instance.
      # 
      # Params
      # ======
      #  Required
      #  --------
      #    * :repository_url => (String)
      #        The URL for the Rational Team Concert repository
      #    * :user_id => (String) 
      #        The user ID under which operations will run
      #    * :password_hash
      #        A hash with either :passwordFile or :password specified
      #        :password - Password for the user ID, a zero-length string by default unless specified
      #        :passwordFile - The password file to use for the connection. You can 
      #           create a suitable password file by invoking the Jazz Build Engine with
      #           the -createPasswordFile argument. You can only specify one password file.
      # 
      def initialize(repository_url, user_id, password_hash={})
        @rtc_args = {
          :repositoryAddress => repository_url,
          :userId => user_id
        }.merge(password_hash)
        
      end
  
      # Returns an object representing a Rational Team Concert Build corresponding
      # to the specified parameters
      # 
      # Params
      # ======
      #  Required
      #  --------
      #    * :uri => (String)
      #        The build's URI
      #    * :uuid => (String) 
      #        The build's internal UUID
      #    * :id
      #        The build's human readable identifier
      # 
      # TODO: We should only require the URI and should be able to derive the other
      # two parameters via a fetch on the URI.
      def resolve_build(uri, uuid, id, type=nil, environment=nil, build_requestor_id=nil)
        return Struct.new(:uri, :uuid, :id, :type, :environment, :build_requestor_id).new(uri, uuid, id, type, environment, build_requestor_id)
      end
  
      # 
      # See BuildSystemProvider#start_activity
      # 
      def start_activity(build, label, parent_activity=nil)
        return start_activity_internal(build, label, parent_activity, false)
      end

      # 
      # See BuildSystemProvider#complete_activity
      #   
      def complete_activity(activity)
    
        task_args = {
          :activityId => activity.id,
          :buildResultUUID => activity.build.uuid, 
        }.merge(@rtc_args)
        
        ant do
          taskdef :name => "completeBuildActivity", :classname => "com.ibm.team.build.ant.task.CompleteBuildActivityTask"
          completeBuildActivity(task_args)
        end
        
      end
  
      # 
      # See BuildSystemProvider#publish_activity
      # 
      def publish_activity(build, label, parent_activity=nil)
        start_activity_internal(build, label, parent_activity, true)
      end
  
      # 
      # See BuildSystemProvider#publish_artifact_link
      # 
      def publish_artifact_link(build, label, url, local_file_path=nil)
        
        task_args = {
          :buildResultUUID => build.uuid,
          :url => url,
          :label => label,
          :localFilePath => local_file_path
        }.merge(@rtc_args)
        
        ant do
          taskdef :name => "artifactLinkPublisher", :classname => "com.ibm.team.build.ant.task.ArtifactLinkPublisherTask"
          artifactLinkPublisher(task_args)
        end
        
      end
      
      # 
      # See BuildSystemProvider#publish_link
      # 
      def publish_link(build, url, label, componentName="Library Entries")
        
        task_args = {
          :buildResultUUID => build.uuid,
          :url => url,
          :label => label,
          :componentName => componentName
        }.merge(@rtc_args)
        
        ant do	
          taskdef :name => "linkPublisher", :classname => "com.ibm.team.build.ant.task.LinkPublisherTask"
          linkPublisher(task_args)
        end
        
      end
      
      def publish_junit_results(build, junit_result_file)
    		
        task_args = {
          :buildResultUUID => build.uuid,
          :filePath => junit_result_file,
          :verbose => "true",
          :failOnError => "true"
        }.merge(@rtc_args)
        
        ant do	
          taskdef :name => "junitResultPublisher", :classname => "com.ibm.team.build.ant.task.JUnitLogPublisherTask"
          junitResultPublisher(task_args)
        end
        
      end
      
      def publish_cloud_resources(build, instance_id, cloud_user_name, cloud_password, cloud_url)
        
        task_args = {
          :buildResultUUID => build.uuid,
          :instanceId => instance_id,
          :cloudUserName => cloud_user_name,
          :cloudUserPwd => cloud_password,
          :cloudRepositoryURL => cloud_url
        }.merge(@rtc_args)
        
        ant do
          taskdef :name => "cloudBuildResourcePublisher", :classname => "com.ibm.devops.rtc.ant.tasks.CloudBuildResultPublisherTask"
          cloudBuildResourcePublisher(task_args)
        end
        
      end
      
      def publish_cloud_logs(build, instance_id, cloud_user_name, cloud_password, cloud_url)
        
        task_args = {
                  :buildResultUUID => build.uuid,
                  :instanceId => instance_id,
                  :cloudUserName => cloud_user_name,
                  :cloudUserPwd => cloud_password,
                  :cloudRepositoryURL => cloud_url
                }.merge(@rtc_args)
                
        ant do
          taskdef :name => "cloudScriptLogPublisher", :classname => "com.ibm.devops.rtc.build.cloudlogs.tasks.CloudLogsPublisherTask"
          cloudScriptLogPublisher(task_args)
        end
        
      end

      # See RtcBuildSystemProvider#publish_log_file      
      def publish_log_file(build, label, log_file, status = 'OK', content_type = 'text/plain')
        task_args = {
          :buildResultUUID => build.uuid,
          :filePath => log_file,
          :label => label,
          :verbose => "true",
          :failOnError => "true",
          :status => status,
          :contentType => content_type
        }.merge(@rtc_args)

        ant do
          taskdef :name => "logPublisher", :classname => "com.ibm.team.build.ant.task.LogPublisherTask"
          logPublisher(task_args)
        end
      end
  
      private
      ##############################################################################
      # Private Methods                                                            #
      ##############################################################################
      
      def start_activity_internal(build, label, parent_activity, auto_complete)
        activity_id = nil

        # Build the argument list that will be used to publish the activity
        
        activity_args = {
          :activityIdProperty => "activityId", 
          :autoComplete => auto_complete,
          :label => label,  
          :buildResultUUID => build.uuid
        }.merge(@rtc_args)
        
        unless parent_activity.nil?
          activity_args[:parentActivityID] = parent_activity.id
        end
        
        ant do
          # Perform the actual ant task and retrieve the id of the activity
          # that was created. This activity_id may be used to create child 
          # activities in the build result UI.
          
          taskdef :name => "startBuildActivity", :classname => "com.ibm.team.build.ant.task.StartBuildActivityTask"
          startBuildActivity(activity_args)
          activity_id = project.getProperty("activityId")
        end
    
        unless auto_complete
          return Struct.new(:id, :build).new(activity_id, build)
        end
      end
  
    end # end RtcProvider
  end # end module Providers
end # end module DevopsBuild
