################################################################################
# Licensed Materials - Property of IBM Copyright IBM Corporation 2012. All Rights Reserved.
# U.S. Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP
# Schedule Contract with IBM Corp.
################################################################################

require "logger"    # Ruby Logging
require 'open-uri'  # Used to get File objects from local or remote files
require 'uri'       # Used to ensure that a string is a URI
require 'timeout'   # Helps throw timeout exceptions
require "tempfile"  # Create a temporary deploy.json file
require "openssl" 

require "devops_deployment_lib/util"
require "devops_deployment_lib/error"
require "devops_deployment_lib/providers/iwd_virtual_system_client" # Ruby client lib to interact with IWD
require "devops_deployment_lib/providers/mock_virtual_system_deployment_manager"

################################################################################
# Class Definition                                                             #
################################################################################
module DevopsDeploymentLib
  
  # A class that helps deploy and monitor the creation of IWD Virtual System 
  # Pattern instances.vsp_client.create_instances
  #
  class VirtualSystemDeploymentManager
    include DevopsDeploymentLib::Util
    
    # Constants
    # =========
    DEFAULT_TIMEOUT = 60
    
    # Class Attributes
    # ================
    attr_reader :vsp_client
    
    @timeout_val
    @log
    
      ################################################################################
      # Class Methods                                                                #
      ################################################################################
      
      # # Description
      # Returns the right VirtualSystemDeploymentManager class.  For local builds use
      # a mock provider that pretends to interact with IWD.
      #  
      # # Arguments
      # - use_mock - (Boolean) If true, return a mock provider
      # 
      def VirtualSystemDeploymentManager.newFor(use_mock, root_server_uri=nil, user_name=nil, password=nil, conf={})
        if use_mock
          return DevopsDeploymentLib::MockVirtualSystemDeploymentManager.new(root_server_uri, user_name, password, conf)
        else
          return DevopsDeploymentLib::VirtualSystemDeploymentManager.new(root_server_uri, user_name, password, conf)
        end
      end
      
      
      ################################################################################
      # Instance Methods
      ################################################################################
    
    
      # Create a new instance of the VirtualSystemDeploymentManager.
      #
      # == Params
      #
      # === Required
      # [root_server_uri (URI String)]  The URI String that points to an IWD server.
      # [user_name (String)]  The user name for the account that will interact with IWD
      # [password [String]]  The password for the account that will interact with IWD
      # 
      # === Optional
      # [:timeout (Integer)]  The max amount of time that the object will wait until it stops
      #                       monitoring the deployment of the Virtual System Pattern instance.
      # [:log_level (Logger Constant)]  The log level that the object will output at.
      #                                 Valid values are:
      #                                 * Logger::FATAL
      #                                 * Logger::ERROR
      #                                 * Logger::WARN
      #                                 * Logger::INFO
      #                                 * Logger::DEBUG
      # [:log_file (String|STDOUT|STDERR)]  The file to output logs to
      #
      def initialize(root_server_uri=nil, user_name=nil, password=nil, conf={})
        conf = {
          :log_level => Logger::DEBUG,
          :log_file => 'http_trace.log',
          :timeout => DEFAULT_TIMEOUT
        }.merge(conf)
      
        @log = Logger.new(conf[:log_file])
        @log.level = conf[:log_level]
      
        @vsp_client = IwdVirtualSystemClient.new(root_server_uri, user_name, password, conf)
        @timeout_val = conf[:timeout]
      end
    
      # Deploy a new instance of a Virtual System Pattern (VSP).
      # This method will wait for for an instance to get created and deployed before returning.
      # Therefore, be sure a suitable timeout is set to accommodate your IWD environment.
      #
      # == Params
      #
      # === Required
      # [application_descriptor (String|URI String)]  A String or String URI to a local or remote file 
      #                                               that contains a JSON formatted ERB template. This
      #                                               content contains the JSON payload to make a successful
      #                                               POST request to create VSP instances on IWD.
      # [environment (Environment)]  An instance of an Environment object tht represents the bindings
      #                              an application will have to the environment it deploys to.
      # == Returns
      # [Hash]  A Hash object representing the data of the new VSP instance that was deployed
      #
      def deploy(application_descriptor, environment)
        unless environment.is_a?(Environment)
          raise ArgumentError, "The 'environment' argument must be an Environment object"
          return
        end
        
        deployment_file = get_deployment_file(application_descriptor, environment)
        instance = @vsp_client.create_instances(deployment_file["content"])
        @vsp_client.ssl_verify_mode = OpenSSL::SSL::VERIFY_NONE
        instance = wait_for_vsp_to_start(instance)
        instance["virtual_machines"] = @vsp_client.describe_instance_virtual_machines(instance["id"])
        #@vsp_client.grant_user_permission_to_instance(instance["id"])
        return instance
      end
      
      # Deploy a new instance of a Virtual System Pattern (VSP).
      # This method will wait for for an instance to get created and deployed before returning.
      # Therefore, be sure a suitable timeout is set to accommodate your IWD environment.
      #
      # == Params
      #
      # === Required
      # [deployment_descriptor]  A String or String URI to a local or remote file 
      #                          that contains a JSON formatted ERB template file. This
      #                          content contains the JSON payload to make a successful
      #                          POST request to create VSP instances on IWD.  The 
      #                          environment bindings must already be set.
      # == Returns
      # [Hash]  A Hash object representing the data of the new VSP instance that was deployed
      #
      def redeploy(deployment_descriptor)
      
        deployment_descriptor_content = deployment_descriptor
      
        if File.file?(deployment_descriptor) or !(deployment_descriptor =~ URI::regexp).nil?
          deployment_descriptor_content = File.open(deployment_descriptor) { |f| f.read }
          @log.debug  "Reading file content"
        end
      
        @log.debug "Deployment descriptor: #{deployment_descriptor}"
        @log.debug "Deployment descriptor content: #{deployment_descriptor_content}"
        
        instance = @vsp_client.create_instances(deployment_descriptor_content)
        @vsp_client.ssl_verify_mode = OpenSSL::SSL::VERIFY_NONE
        instance = wait_for_vsp_to_start(instance)
        instance["virtual_machines"] = @vsp_client.describe_instance_virtual_machines(instance["id"])
        #@vsp_client.grant_user_permission_to_instance(instance["id"])
        return instance
      end
      
      ##########
      # Create a deployment file (JSON format), but don't save it anywhere.
      #
      # == Params
      # === Required
      # [application_descriptor (String|URI String)]  A String or String URI to a local or remote file 
      #                                               that contains a JSON formatted ERB template. This
      #                                               content contains the JSON payload to make a successful
      #                                               POST request to create VSP instances on IWD.
      # [environment (Environment)]  An instance of an Environment object that represents the bindings
      #                              an application will have to the environment it deploys to.
      #  == Returns
      # [deployment_file] (Hash) - A Hash object representing the deployment file
      #      deployment_file["content"] = the contents (JSON) of the temporary file
      #
      def get_deployment_file(application_descriptor, environment)
        
        application_descriptor_content = application_descriptor
      
        if File.file?(application_descriptor) or !(application_descriptor =~ URI::regexp).nil?
          application_descriptor_content = open(application_descriptor) { |f| f.read }
        end
        
        instance_attributes = evaluate_template(application_descriptor_content, environment)
        ia_json = create_hash_from_json(instance_attributes)
        
        deployment_file = Hash.new()
        deployment_file["content"] = ia_json.to_json
        
        return deployment_file
      end
            
      ##########
      # Create a deployment file (JSON format), and save it to a temporary folder.  The folder will be created if necessary.
      #
      # == Params
      # === Required
      # [application_descriptor (String|URI String)]  A String or String URI to a local or remote file 
      #                                               that contains a JSON formatted ERB template. This
      #                                               content contains the JSON payload to make a successful
      #                                               POST request to create VSP instances on IWD.
      # [environment (Environment)]  An instance of an Environment object tht represents the bindings
      #                              an application will have to the environment it deploys to.
      #  == Returns
      # [deployment_file] (Hash) - A Hash object representing the deployment file
      #      deployment_file["content"] = the contents (JSON) of the temporary file
      #      deployment_file["path"] = the full pathname of the file (note that this will be deleted when the process exits)
      #      deployment_file["basename"] = the basename of the file
      #
      def create_deployment_file(application_descriptor, environment)

        deployment_file = get_deployment_file(application_descriptor, environment)
        # Tempfile class handles temp location on Linux vs. Windows
        tempfile = Tempfile.new("deploy.json")
        tempfile.write(deployment_file["content"])
        tempfile.close
        deployment_file["path"] = tempfile.path
        deployment_file["basename"] = File.basename(tempfile.path)
        
        return deployment_file
      end
      
      # Set the ssl_verify_mode on @vps_client.  This is a pass-through method
      #
      # == Params
      # === Required
      # [mode] (Boolean) True means to do ssl verification
      #
      def set_ssl_verify_mode(mode)
        @vsp_client.ssl_verify_mode = mode
      end
    
      private
      ################################################################################
      # Private Methods                                                              #
      ################################################################################
    
      # Wait for the Virtual System Pattern instance to start by polling the resource.
      #
      # == Returns
      # [Hash]  A hash object representing the instance that was created on IWD
      #
      def wait_for_vsp_to_start(instance)
        servercodes_yaml = YAML::load(File.open(File.join(File.dirname(File.expand_path(__FILE__)), 'server_codes.yaml')))
          @log.debug servercodes_yaml
        Timeout::timeout(@timeout_val) do
          while ( !instance["currentstatus"].eql?("RM01006") ) #RM01006 code = Started
            
            #If instance has started
            if(instance["currentstatus"].eql?("RM01006"))
               instance = @vsp_client.describe_instance(instance["id"])
               @log.debug instance["currentmessage_text"]
               break  
            end
            
            if !servercodes_yaml.nil?
              if servercodes_yaml["ERROR"].include?(instance["currentstatus"])
                instance = @vsp_client.describe_instance(instance["id"])
                @log.debug instance["currentmessage_text"]
                break
              end
            else
                @log.debug 'error loading server_codes.yaml'
            end
            
            sleep 30
            instance = @vsp_client.describe_instance(instance["id"])
            @log.debug "Instance: #{instance.inspect}"
            
          end
        end
        #instance should be started now, if not raise an exception
        if(!instance["currentstatus"].eql?("RM01006"))
          instance = @vsp_client.describe_instance(instance["id"])
          ex = Exceptions::ProvisioningError.new(instance["currentstatus_text"])
          ex.details = instance
          raise ex   
          @log.debug instance["currentmessage_text"]
        end
        return instance
      end
    
  end # end VirtualSystemDeploymentManager class
end # end DevopsDeploymentLib module
