################################################################################
# Licensed Materials - Property of IBM Copyright IBM Corporation 2012. All Rights Reserved.
# U.S. Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP
# Schedule Contract with IBM Corp.
################################################################################

require "tempfile"
require 'zip/zip'
require 'zip/zipfilesystem'
require 'devops_library/zipentry.rb'


module DevopsLibrary

  ##########
  # Class Definition
  # The ChefPackager creates:
  #  - a .zip file containing the chef cookbooks
  # AND ALSO updates a metadata hash to have entries for file names and paths
  class ChefPackager

    ##########
    # Create a .zip file containing the chef cookbooks
    # Before calling this method, put these values into the metadata hash:
    #   :chef_root - the path to the cookbooks to package; must have a "cookbooks subdirectory"
    #   :basename (optional)- the application name of the chef package, in the asset library
    #       if basename is not provided, it will be calculated from the name of the chef_root
    #       directory
    # This method will update the metadata hash with these values:
    #   :infrastructure_path - full path (including filename) to the temporary .zip file 
    #       containing the chef cookbooks.  This filename includes a unique ID to avoid 
    #       collisions in the temporary directory.
    #   :infrastructure_filename - just the filename (no path) of the final .zip file in the 
    #       asset library.  This filename does not include the unique ID.
    # This method returns the same hash that was passed in (this makes calling code simpler)
    def create_package(metadata={})
      
      validate_metadata(metadata)
      create_package_file(metadata)
      return metadata
      
    end


    ################################################################################
    # Private Methods
    private


    ##########
    # Create a zip file and update the metadata hash
    def create_package_file(metadata)
      
      root_folder = metadata[:chef_root]

      # Create the .zip file
      metadata[:basename] ||= File.basename(root_folder)
      # Tempfile class handles temp location on Linux vs. Windows
      base = Tempfile.new(metadata[:basename]).path
      zip_path = "#{base}.zip"
      
      base_name_length = root_folder.length + 1
      Dir.chdir(root_folder) do
        Zip::ZipFile.open(zip_path, Zip::ZipFile::CREATE) do |zipfile|
          Dir["#{root_folder}/**/**"].each do |file|
            begin  
              entry = file[base_name_length..file.length]  
              new_entry = DevopsLibrary::ZipEntry.new(zipfile.to_s, entry.to_s)
              zipfile.add(new_entry, file)
            rescue  Exception => e
              raise "Error occured : #{file} #{e}"
            end
          end
        end
      end


      # Add to the metadata
      metadata[:infrastructure_path] = zip_path
      metadata[:infrastructure_filename] = "#{metadata[:basename]}.zip"
      
    end


    ##########
    # Validate the metadata hash contains the required entries
    def validate_metadata(metadata)

      # make sure the root folder is a directory
      root_folder = metadata[:chef_root]
      if !File.directory?(root_folder)
        raise ArgumentError, ":chef_root must be a directory; the value passed in for :chef_root was #{root_folder}"
      end
      
      # make sure the root folder has a cookbooks subdirectory
      if !File.directory?(File.join(root_folder, "cookbooks"))
        raise ArgumentError, ":chef_root must minimally include a 'cookbooks' directory; the value passed in for :chef_root was #{root_folder}"
      end
      
    end



  end # End class
end # end module
