require 'weaver/cli/weaver_cli_plugin'
require 'weaver/chef/chef_node_inspector'
require 'weaver/chef/chef_automation_handler'
require 'fileutils'

module Weaver
  module Cli
    module Plugins
      # \Easyenv plugin for the Weaver CLI
      #
      # Syntax: <tt>weaver easyenv [task] [options]</tt>
      #
      class EasyenvCliPlugin < Weaver::Cli::WeaverCliPlugin

        # Special non-Thor extension to provide help for the arguments of the commands
        ARG_DEFS = {
          'ENVIRONMENT_FILE' => 'A .weaver file defining an environment',
        }
      
        namespace :easyenv
        namespace_description "EasyEnv related tasks"

        desc "import ROLE_FILE", "Create a Weaver description of a Chef Role file (*_role.rb)"
        method_option :cookbooks, :type => :string, :desc => "Colon-separated folders named 'chef' containing Chef setup."
        method_option :loc, :type => :string, :aliases => "-l", :desc => "Location for the resulting .weaver file."
        # Create a \Weaver description of a \Chef role file (*_role.rb)
        # * +file+ - \Chef role file (*_role.rb)
        # * +cookbooks+ - Colon-separated folders named 'chef' containing \Chef setup.
        # * +loc+ - Location for the resulting .weaver file.
        def import(file = nil)
          setup_logging
          collect_properties

          # Limitation in Thor and JRuby so we have to missing parameters by assigning defaults and checking for them
          raise Thor::Error.new(%Q{"import" was called incorrectly. Call as "import ROLE_FILE".}) unless file

          # 'chef_path' should be something like "~/src/eclipse/devops/com.ibm.devops.sample.infrastructure/chef:~/src/eclipse/devops/com.ibm.devops.common.infrastructure/chef"
          chef_path = options[:cookbooks] || ENV['CHEF_SETUP_PATHS']
          say_status("Error", "Parameter 'cookbooks' or environment variable 'CHEF_SETUP_PATHS' not set.", :red) && exit(1) unless chef_path

          unless chef_path.nil?
            chef_path.split(File::PATH_SEPARATOR).each do |dirname|
              unless File.directory?(dirname)
                say_status("Error", "Directory #{dirname} does not exist", :red)
              else
                say_status("Error", "#{dirname} is not a Chef directory (does not contain 'cookbooks' or 'roles' subdirectories)", :red) unless
                File.directory?(File.expand_path('cookbooks', dirname)) || File.directory?(File.expand_path('roles', dirname))
              end
            end
          end

          inspector = Weaver::Chef::ChefNodeInspector.new(chef_path.split(File::PATH_SEPARATOR)) do |level, msg|
            report_message("Importing", level, msg)
          end

          dir = options[:loc] || '.' # where to store it

          begin
            model = inspector.model_role(file, dir)
            extension = ".rb"
          rescue WeaverError => we
            say_status("Error", "Chef introspection failed. Original message: '#{we.to_s}'", :red)
            exit(1)
          end

          name = File.basename(file, extension)
          full_filename = filename = name + Weaver::Core::WEAVER_SUFFIX

          if options[:loc]
            full_filename = File.join(dir, full_filename)
          end

          say_status("Saving", "Role definition '#{filename}' in '#{dir}'", :green)

          # creating a backup copy if file already exists
          if File.exists?(full_filename)
            say_status("Creating:", "File '#{filename}' already exists. Creating backup '#{filename}.saved'", :green)
            FileUtils.copy(full_filename, "#{full_filename}.saved")
          end

          FileUtils.mkdir_p(File.dirname(full_filename)) unless File.directory?(File.dirname(full_filename))
          file = File.open(full_filename, 'w')
          File.open(file.path, 'w') do |f|
            f.write( model.to_weaver )
          end

        end

        desc "nodes ENVIRONMENT_FILE", "Generates the Chef node files (for chef-solo) for a given Weaver environment"
        method_option :loc, :type => :string, :aliases => "-l", :desc => "Location for the resulting Chef node file."

        # Create a \Weaver description of a \Chef role file (*_role.rb)
        # * +environment_file+ - \Weaver environment file
        def nodes(environment_file = nil)
          setup_logging
          collect_properties
          
          # Limitation in Thor and JRuby so we have to missing parameters by assigning defaults and checking for them
          raise Thor::Error.new(%Q{"nodes" was called incorrectly. Call as "nodes ENVIRONMENT_FILE".}) unless environment_file
          
          if options[:loc]
            dir = File.expand_path(options[:loc])                    
          else 
            dir = "."
          end
          
          if File.exists?(environment_file) && File.directory?(environment_file)
            say_status("Error", "File '#{environment_file}' is a directory.", :red)
            exit(1) 
          end
          if !File.exists?(environment_file) 
            say_status("Error", "File '#{environment_file}' does not exist.", :red)
            exit(1)
          end          
          if !File.directory?(dir)
            say_status("Error", "Directory '#{dir}' is not a valid directory.", :red)
            exit(1)
          end
          
          handler = Weaver::Chef::ChefAutomationHandler.new() { |level, msg| report_message("Generating", level, msg) }            
          begin
            # This should use Environment here as it implements the logic already but this CLI function would require licensing then.
            root = Weaver::Dsl::WeaverParser.new().parse(environment_file)
            topology = root.topologies.values.to_a[0]
            if root.topologies.size > 1
              @@logger.warn "#{environment_topology_file} contains #{@root.topologies.size} top-level topologies, using #{@topology.id.inspect}"
            end
            if topology.nil?
              @@logger.error "No topology in #{environment_topology_file.inspect}"
            end
            topology = Weaver::Transforms::Fold.fold(topology)

            handler.generate_node_files(topology, dir)
            say_status("Done", "chef-solo node files generated successfully.", :green)

          rescue WeaverError => we
            say_status("Error", "chef-solo node file generation failed. Original message: '#{we.to_s}'", :red)
            exit(1)
          end
        end

        no_tasks do
          # @private
          # :stopdoc:
          # Report a status message running a command
          def report_message(status, level, msg)
            if level == Logger::ERROR || level == Logger::FATAL
              say_status(status, msg, :red)
            elsif level == Logger::WARN
              say_status(status, msg, :yellow)
            else
              say_status(status, msg, :green)
            end
          end
          # :startdoc:
        end

      end
    end

  end
end