require 'set'
require 'weaver/utils/reversing_hash'
require 'weaver/validation/validation_error'
require 'weaver/validation/problem_status'

module Weaver
  module Core
    # Implements basic functionality for a model element in our component model.
    #
    # Author:: Florian Rosenberg
    # Author:: Ed Snible
    class Abstract
      include Weaver::Validation

      # :stopdoc:
      # Unlike, e.g. @@validate_handlers, which is shared between classes the following will be unique per class
      class << self
        # Methods to +initialize_copy()+ the children of this entity.  For example, the Node subclass needs +:initialize_copy_nodes+
        # to deeply copy contained Nodes.  These methods shouldn't copy references, only contained entities. 
        attr_accessor :initialize_copy_handlers
        
        # Methods to validate the children of this entity.  For example, the NodeContainerMixin mixin uses +:validate_nodes+ which
        # if invoked with +send+ validates the contained Nodes.
        attr_accessor :validate_handlers
        
        # Methods to return the children of this entity.  For example, the Node subclass has a child_container +:nodes+ which
        # if invoked with +send+ returns children of that type.
        attr_accessor :child_enumerators
        
        # Methods to return freshen references from this entity to other Bases during operations such as Duplicate.duplicate()
        # which make deep copies.  (References are ordinarily not updated during +initialize_copy()+. 
        attr_accessor :reference_fresheners
        
        # Methods to return the non-containment children of this entity.  For example, the Node subclass has a child_container +:alias_nodes+ which
        # if invoked with +send+ returns children of that type.
        attr_accessor :child_containers
        
        # Methods to return non-containment references from this entity to other Bases.  For example, NodeContainerMixin puts
        # +:alias_nodes+ here so that developers inspecting the model can inspect node aliases.
        attr_accessor :reference_enumerators
      end
      # :startdoc:
      
      # Called implicitly by the interpreter during subclassing; this method makes sure the containment and reference handlers respect
      # subclassing.  Never call this directly; Ruby calls it for you.
      def self.inherited(subclazz)
        subclazz.initialize_copy_handlers = [] if subclazz.initialize_copy_handlers.nil?
        subclazz.initialize_copy_handlers.concat(initialize_copy_handlers) unless initialize_copy_handlers.nil?
        
        subclazz.validate_handlers = [] if subclazz.validate_handlers.nil?
        subclazz.validate_handlers.concat(validate_handlers) unless validate_handlers.nil?
        
        subclazz.child_enumerators = [] if subclazz.child_enumerators.nil?
        subclazz.child_enumerators.concat(child_enumerators) unless child_enumerators.nil?
        
        subclazz.reference_fresheners = [] if subclazz.reference_fresheners.nil?
        subclazz.reference_fresheners.concat(reference_fresheners) unless reference_fresheners.nil?
        
        subclazz.child_containers = [] if subclazz.child_containers.nil?
        subclazz.child_containers.concat(child_containers) unless child_containers.nil?
        
        subclazz.reference_enumerators = [] if subclazz.reference_enumerators.nil?
        subclazz.reference_enumerators.concat(reference_enumerators) unless reference_enumerators.nil?
      end
      
      # The id (Symbol) of a model element (required).
      attr_accessor :id

      # The Weaver type of a model element; a Symbol such as :node, :component
      attr_accessor :type

      # Hash of id to Attribute (an Attribute is MetaData and a value)
      attr_accessor :amd
      
      # The enclosing Weaver block; it is a Base
      attr_accessor :scope

      # Array of Arrays which take the form [Symbol, *params]
      attr_reader :validators
      
      # Where (in a .weaver source file) was this entity defined
      attr_accessor :weaver_origin
      
      # Comments to add when calling :to_weaver to render this entity.  These comments will be rendered before the entity.
      attr_accessor :prefix_comments
      
      # Comments to add when calling :to_weaver to render this entity.  These comments will be rendered within the entity.
      attr_accessor :inner_comments
      
      # ReversingHash representing reverse of group.member
      attr_reader :member_of
      
      # Represents the 'base type' of an entity
      attr_accessor :provenance
      
      # Hash representing +use+ processing that is not yet complete (used for generating .weaver files from the model)
      # The keys are desired IDs, the values are imported model objects
      attr_reader :unresolved_refinements
      
      # Set of symbols associated with this model object
      attr_reader :tags
      
      # * +id+ Symbol representing the entity
      # * +parent+ The enclosing Base
      def initialize(id, parent)
        
        if !parent.nil? && !parent.kind_of?(Base)
          raise WeaverError, "'parent' must be a Weaver Model Element"
        end
        
        if id.instance_of? Hash then
          key = id.keys[0]
          @id = key
          @type = id[key]
        else
          @id = id
        end
        @amd = {}
        @validators = []
        @scope = parent
        
        raise "@id = #{@id}, a #{@id.class}" unless @id.kind_of?(Symbol) || @id.kind_of?(String) || @id.nil?
        
#        @implicit_attributes = {}
        
        # Don't create :name and :description on Attribute.  TODO Refactor Attribute so it isn't a Base
        unless kind_of?(Attribute)
          @amd[:name] = Attribute.new(:name, self, true)
          @amd[:name].modeled = true
          @amd[:name].implicit = true
          @amd[:name].required = false
          @amd[:description] = Attribute.new(:description, self, true)
          @amd[:description].modeled = true
          @amd[:description].implicit = true
          @amd[:description].required = false
        end
        
        @prefix_comments = []
        @inner_comments = []
        
        @member_of = Weaver::Utils::ReversingHash.new(self, :members)
        @unresolved_refinements = {}
        @tags = Set.new
      end

      # All Weaver entities define an +initialize_copy+ to deeply copy their children.
      def initialize_copy(other)
        @member_of = Weaver::Utils::ReversingHash.new(self, :members)     # We will rebuild this later
        
        super
        
        @amd = {}
        other.amd.each { |id, attribute|
          dup = attribute.clone()
          dup.scope = self
          @amd[id] = dup
        }
        
#        @implicit_attributes = {}
#        other.implicit_attributes.each { |id, attribute|
#          dup = attribute.clone()
#          dup.scope = self
#          @implicit_attributes[id] = dup 
#        }
        
        @validators = other.validators.clone() 
        @inner_comments  = other.inner_comments.clone()
        @prefix_comments = other.prefix_comments.clone()
        @tags = other.tags.clone
        
        self.class.initialize_copy_handlers.each do |func| 
          send(func, other)
        end
      end
      
      # (Weaver overrides <code>inspect</code> because the built-in inspect was following reference cycles and generating large
      # an infinite-sized inspections.)
      def inspect
        return to_s
      end
      
      # Find the attribute +sym+ and clear its +implicit+ flag.
      # @see Attribute#implicit?
      def make_attribute_explicit(sym)
        # Promote implicit attribute to regular attribute
#        attr = @implicit_attributes.delete(sym).clone()
#        attr.scope = self
#        @amd[sym] = attr
        amd[sym].implicit = false
      end
      
      # If 'name' is a late_binding and not evaluable we will have problems during parsing
      # with debugging statements; this makes name() never nil
      def name
        safe_evaluate(:name)
      end
      
      # If +description+ is a late_binding and not evaluable we will have problems during parsing
      # with debugging statements; this method evaluates it so that makes description() is never +nil+.
      def description
        safe_evaluate(:description)
      end

      # If we need a value but DON'T want to evaluate late bindings use this instead
      def safe_evaluate(sym)
        attr = amd[sym]
        raise WeaverError.new("Unknown attribute :#{sym} on #{fully_qualified_id}") if attr.nil?
        begin
          return attr.value if attr.unresolved_value.kind_of?(LateBinding)
        rescue TypeError => e
          return "late_binding {...}"
        end
        return attr.value
      end

      # Registers an initialize_copy handler. The initialize_copy() method calls all registered handlers
      # Using this style we can mix in containment of Weaver types.
      def self.register_initialize_copy_handler(element)
        if self.initialize_copy_handlers.member?(element)
          raise WeaverError.new("Duplicate handler #{element} being added to #{self}")
        end
        
        self.initialize_copy_handlers.push(element)
      end
      
      # Registers an validate_collect handler. The validate_collect() method calls all registered handlers
      # Using this style we can mix in containment of Weaver types.
      def self.register_validate_handler(element)
        if self.validate_handlers.member?(element)
          raise WeaverError.new("Duplicate handler #{element} being added to #{self}")
        end
        
        self.validate_handlers.push(element)
      end
      
      # Registers a child Weaver entity enumerator. The child_entities() method calls all registered enumerators
      # Using this style we can mix in containment of Weaver types.
      # Note that due to how Base.rename() is implemented the child_enumerators should not return a temporary, but
      # instead return a live object, because we use this not just to enumerate the children but the rename them.
      def self.register_child_enumerator(element)
        if self.child_enumerators.member?(element)
          raise WeaverError.new("Duplicate handler #{element} being added to #{self}")
        end
        
        self.child_enumerators.push(element)
      end
      
      # Registers a container method for child elements
      def self.register_child_container(element)
        if self.child_containers.member?(element)
          raise WeaverError.new("Duplicate container handler #{element} being added to #{self}")
        end
        
        self.child_containers.push(element)
      end
      
      # Registers a 'reference freshener'; this will be called by operations such as Duplicate.duplicate after contained
      # objects have changed so that references to those objects can be changed.
      def self.register_reference_freshener(element)
        if self.reference_fresheners.member?(element)
          raise WeaverError.new("Duplicate handler #{element} being added to #{self}")
        end
        
        self.reference_fresheners.push(element)
      end
  
      # Registers a reference Weaver entity enumerator. Like a child enumerator except by reference, not by value
      def self.register_reference_enumerator(element)
        if self.reference_enumerators.member?(element)
          raise WeaverError.new("Duplicate reference enumerator handler #{element} being added to #{self}")
        end
        
        self.reference_enumerators.push(element)
      end

      # Keep the old respond-to, so we can tell the difference between Ruby and Weaver responses
      alias_method :original_respond_to?, :respond_to?
      
      # Let developers known which Weaver things can be manipulated.
      # @see #method_missing
      def respond_to?(sym, include_private=false)
        # NOTE: In Ruby 1.9.x it is better to define respond_to_missing?() than respond_to?(); we do this to be compatible.
        
        if sym.to_s[-1,1] == '='        
          # We are being asked to set an attribute; do we have an attribute by that name?
          return @amd.has_key?(sym.to_s.chop.to_sym)
        end
        
        # Do we have an attribute or child entity by that name?
        return true if child_entities.member?(sym)
        
        super
      end
      
      # :stopdoc:
      def method_missing(sym, *arguments, &block)

        # allow attributes to be referenced by its name
        if @amd.key?(sym)
          return amd[sym].value
        elsif @amd[sym.to_s] != nil
          returm amd[sym.to_s].value
#        elsif @implicit_attributes.key?(sym)
#          return @implicit_attributes[sym].value
        elsif sym.to_s[-1,1] == '='        
          # We are being asked to set, do we have an attribute by that name?
          rsym = sym.to_s.chop.to_sym                
          if amd.key?(rsym)
            if amd[rsym].implicit?
              make_attribute_explicit(rsym)
            else
              unless amd[rsym].mutable
                raise WeaverError.new("#{rsym} is not mutable and cannot be changed")
              end
            end
            amd[rsym].value = arguments[0]
#          elsif @implicit_attributes.key?(rsym)
#            make_attribute_explicit(rsym)
#            amd[rsym].value = arguments[0]
          else
            # No attribute by that name
            super
          end
        else
          child = child_entities[sym]
          return child unless child.nil?

          # We have an undefined variable.  Can we detect the caller?
          @@logger.error "Unknown `#{sym}' on #{self.class} executing #{caller[0]}"          
                    
          super
        end
      end
      # :startdoc:
      

      # Returns <code>true</code> if +another_base+ is equal by value.
      def ==(another_base)
        if !another_base.kind_of?(Base)
          return false
        end
        
        raise WeaverError, "id must be a Symbol or String but it is a #{id.class} and self is a #{self.class} [#{self.to_s}]" unless @id.kind_of?(Symbol) || @id.kind_of?(String)

        retval = @id == another_base.id && @type == another_base.type && 
         @amd == another_base.amd # && child_entities == another_base.child_entities
        raise WeaverError.new("Comparison problem, retval is a #{retval.class}") unless retval.kind_of?(TrueClass) || retval.kind_of?(FalseClass)
        return retval
      end
      
      # Model elements enclosed by this one, both by containment and by reference
      def child_entities
        retval = amd.clone
        
        self.class.child_enumerators.each do |func|
          retval.merge!(send(func))
        end

        return retval
      end
      
      # Returns the entities that are contained by value but not those contained by
      # reference.
      # @see #child_entities
      # @see #child_references
      def contained_entities
        aliases = {}
        self.class.reference_enumerators.each do |func|
          aliases.merge!(send(func))
        end
        
        return child_entities.delete_if { |id, child| aliases.member?(id) }
      end
      
      # Returns the entities that are contained by reference
      # @see #child_entities
      # @see #contained_references
      def child_references
        retval = {}
          
        self.class.reference_enumerators.each do |func|
          retval.merge!(send(func))
        end

        return retval
      end
      
      # RECURSIVE
      # This method updates non-containment links to other Weaver entities after an operation like Duplicate.duplicate.
      # Ruby's initialize_copy method is used in Weaver to make deep copies of containment links, but references should
      # be updated by registered referenence 'fresheners'.
      def update_references(old_to_new)
        self.class.reference_fresheners.each do |func|
          send(func, old_to_new) if self.respond_to?(func)
        end
        
        contained_entities.each { |id, child| child.update_references(old_to_new) } 
      end
      
      # Rename a contained Weaver entity or attribute
      # * *Args*    :
      #   - +entity+ -> a Weaver entity that is a direct or indirect child of model
      #   - +new_id+ -> a Symbol which is to be the new id of entity
      #
      def rename(entity, new_id)
        if child_entities.has_key?(new_id)
          raise WeaverError.new("Cannot rename #{entity.id.inspect} because #{fully_qualified_id} already has #{new_id.inspect}")        
        end
        
        # It isn't sufficient to tell the entity its new ID; its container also needs to know.
        # The container supplies child_entities() to resolve names but this isn't a two-way data
        # structure; modifying it won't help.  Instead we must query @amd and @@child_enumerators.

        if amd[entity.id].equal?(entity)
          amd.delete(entity.id)
          entity.id = new_id
          amd[new_id] = entity
          return entity
        end
        
        # @@child_enumerators contains symbols for the getter, rather than creating a separate
        # data structure for the setter we will append '=' to the getter.
        
        self.class.child_enumerators.each do |func|
          candidate_array = send func
          if entity.equal?(candidate_array[entity.id])
            candidate_array.delete(entity.id)
            entity.id = new_id
            candidate_array.store(new_id, entity)
            
            unless entity.equal?((send func)[new_id]) 
              raise WeaverError.new("Rename failed, child_enumerator for #{entity.id} doesn't return persistent Array")
            end
        
            return entity
          end 
        end

        raise WeaverError.new("No entity :#{entity.id} to rename in #{type} :#{id}")        
      end

      # Creates an alias for a Weaver entity
      # * *Args*    :
      #   - +entity+ -> a Weaver entity
      #   - +new_id+ -> a Symbol which is to be the new id of entity
      #
      def alias_of(entity, new_id)
        
        # @@child_enumerators contains symbols for the getter, rather than creating a separate
        # data structure for the setter we will append '=' to the getter.

        self.class.child_containers.each do |func| 
          accepted = send func, new_id, entity
          return if accepted
        end
      
        raise WeaverError.new("No container for #{entity.type} :#{entity.id} in #{type} :#{id}")        
      end

      # :stopdoc:
      def to_s
        return "#{type} (:#{id.to_s}) { attributes: #{amd.keys.inspect}, #{child_entities.size} child entities }"
      end
      # :startdoc:

      # Calls validate_collect() and throws an exception if any validation messages are generated
      # * +desired_phases+ an optional Array which may be +[:predeployment]+
      # @see #validate_collect
      def validate(desired_phases = nil)
        out_messages = []
#        def out_messages.push(obj)
#          raise "Can't push #{obj.class}" unless obj.kind_of?(Weaver::Validation::ProblemStatus)
#          super
#        end
#        def out_messages.concat(other_ary)
#          raise "Can't concat #{other_ary}" unless other_ary.detect { |obj| !obj.kind_of?(Weaver::Validation::ProblemStatus) }.nil?
#          super
#        end
        
        root_entity.unreported_messages.clear unless root_entity.nil?
        validate_collect(out_messages, desired_phases)
        
        out_messages.concat(root_entity.unreported_messages) unless root_entity.nil?
        if out_messages.size > 0
          raise Weaver::Validation::ValidationError.new(out_messages.to_s)
        end
      end
      
      # Validates the entity and its direct and indirect contained children.
      # * +out_messages+ ProblemStatus will be pushed onto this for any validation message
      # * +desired_phases+ optional Array which may be +[:predeployment]:
      # * +excluded_phases+ optional Array which may be +[:predefined]: DEPRECATED
      # @see #validate
      def validate_collect(out_messages, desired_phases = nil, excluded_phases = nil)
        
        # Let validators look up the current validation phase
        Thread.current[:weaver_validation_phases] = desired_phases || []
          
        freshen_member_of
        
        return if tags.member?(:novalidate)
        
        #puts "inside base.validate_collect... desired_phases = #{desired_phases.inspect}"
        if id.kind_of?(Symbol) && id.to_s =~ /^[[:upper:]]/ && !self.kind_of?(Attribute)
          report(Logger::ERROR, "IDs must begin with lower-case letter: #{id.inspect}", self)
        end
        
        # Attribute names (the names passed to the cloud layer) must be unique
        attr_names = @amd.values.select { |attr| attr.name != nil || attr.group != nil }.map { |named_attr| "#{named_attr.group}.#{named_attr.name}" }
        if attr_names.size != attr_names.uniq.size
          duplicated_names = Set.new(attr_names.select { |name| attr_names.count(name) > 1 })
          @amd.values.each do |attr|
            if duplicated_names.member?("#{attr.group}.#{attr.name}")
              attr.add_validation_failure("The name '#{attr.group}#{'.' unless attr.group.nil?}#{attr.name}' is not unique") 
            end
          end 
        end
        
        @amd.each { |attributeid, attribute| attribute.validate_collect(out_messages, desired_phases) }
          
        self.class.validate_handlers.each do |func| 
          send(func, out_messages, desired_phases)
        end

        @member_of.each do |id, group|
          out_messages.push ProblemStatus.new(Logger::WARN, "Unknown group :#{id}", self) if group.nil?
        end

        # Do we have a library of validators?
        if @validators.length > 0 && root_entity.nil?
          out_messages.push ProblemStatus.new(Logger::WARN, "No validator context", self)
          return
        end
           
        @validators.each { |symb, *params| call_validator(symb, params, out_messages, desired_phases) }
        
        unless root_entity.nil?  
          out_messages.concat(root_entity.unreported_messages)
          root_entity.unreported_messages.clear
        end
      end
      
      # Call a named validator with parameters
      # * +symb+ the validator symbol, e.g. +:equals+
      # * +params+ Array of parameters
      # * +out_messages+ an Array in which ProblemStatus will be pushed
      # * +desired_phases+ optional Array containing +[:predeployment]+
      def call_validator(symb, params, out_messages, desired_phases = nil)
          
        # Do we have the validator?
        validator = root_entity.named_validators[symb]
        if validator.nil?
          out_messages.push ProblemStatus.new(Logger::ERROR, "Unknown validator :#{symb}", self)
          return
        end
        
        # Does the validator talk to our kind of entity?
        if validator.type != :any
          out_messages.push ProblemStatus.new(Logger::ERROR, ":#{symb} can only validate :#{validator.type}s", self) unless type == validator.type 
        end
        
        valproc = validator.proc
        raise WeaverError.new(":#{symb} has no Proc") unless valproc.kind_of?(Proc)

        # Handle any late bindings in the parameters          
        params = params.map { |v| v.kind_of?(LateBinding) ? v.weaver_eval(self) : v } 
        
        number_of_mandatory_arguments = valproc.arity < 0 ? -valproc.arity-1 : valproc.arity
        if params.length < number_of_mandatory_arguments
          weaver_file = valproc.source_location[0]
          weaver_line = valproc.source_location[1]
          out_messages.push ProblemStatus.new(Logger::ERROR, "#{weaver_file}:#{weaver_line} validator :#{symb} needs #{number_of_mandatory_arguments} but only #{params.length} parameters supplied (#{params}", self)
        else

          if (desired_phases.nil? || desired_phases.length == 0) 
            ## User wants universal validators only
            if (!validator.phase.nil?) && validator.phase != :any && validator.phase != :basic
              # This validator is not universal, i.e., it has a phase; thus, it needs to be skipped.
              @@logger.debug "Skipping validator :#{validator.inspect}. Validator phase = #{validator.phase}. User wants universal validators only."
              return
            end
          elsif (!validator.phase.nil?) && validator.phase != :any && (!desired_phases.member?(validator.phase))
            ## This validator is not universal and does not belong to the desired phases. Skip it.
            @@logger.debug "Skipping validator :#{validator.inspect}. Validator phase = #{validator.phase}. Desired phases = #{desired_phases.inspect}"
            return
          end
          
          @@logger.debug "Calling validator :#{symb}(#{params.inspect}) validator phase: #{validator.phase}"
          begin
            instance_exec(*params, &valproc)
          rescue Weaver::Validation::ValidationError => theException
            weaver_file = valproc.source_location[0]
            weaver_line = valproc.source_location[1]
            context_id = theException.context.nil? ? nil : theException.context.id
            origin = theException.context.nil? ? nil : theException.context.weaver_origin
            level = theException.level
            out_messages.push Weaver::Validation::ProblemStatus.new(Logger::ERROR, "#{weaver_file}:#{weaver_line} validator failure: #{theException.message}; context=#{context_id}", self, origin)
          end
        end
      end
    
      # Add a validator.  For example, to add <code>{ equals(3) }</code>, call with <code>add_validator(:equals, 3)</code>. 
      def add_validator(validator_symbol, *params)
        raise WeaverError.new("validator_symbol must be a Symbol but it is #{validator_symbol} (a #{validator_symbol.class})") unless validator_symbol.kind_of?(Symbol)
        @validators.push([validator_symbol].concat(params))
        
        # We don't type-check the validator application here (e.g. => :attribute validators can only
        # be attached to Attributes) because this Base might not yet have a Root initialized with
        # the validator library.
      end

      # The error message will appear when the entity is validated
      # * *Args*    :
      #   - +msg+ -> Message to display to the user
      #   - +origin+ -> optional String filename:lineno if different than the containing Entity's most recent declaration of origin
      def add_validation_failure(msg, origin=nil)
        add_validator(:always_fail, msg, origin)
      end
      
      # Report a validation failure.  (Mostly used by validators and rules.)
      # * *Args*    :
      #   - +level+ -> a Logger level; Logger::ERROR is usual
      #   - +message+ -> an untranslated message
      #   - +context+ -> a Weaver entity associated with the message
      #   - +exception+ -> optional exception
      #   - +type+ -> optional special type (e.g. an IWD problem)
      #   - +description+ -> optional additional details
      #   - +resolutions+ -> optional array of Resolutions with evaluated strings
      #   - +origin+ -> a file:lineno String if different than the line declaring 'context'
      def report(level, message, context, exception = nil, type = nil, description=nil, resolutions=[], origin=nil)
        # Don't accept reports from rules for entities with :novalidate
        return if context.tags.member?(:novalidate)
        
        return root_entity.report(level, message, context, exception, type, description, resolutions, origin)
      end
      
      # The Root which contains the \Weaver document
      def root_entity
        entity = self
        while !entity.nil? && !entity.kind_of?(Root)
          entity = entity.scope
        end
        return entity
      end

      # Look for a direct child with a particular ID; if no child of that ID exists creates one using clazz.
      # * +id+ a Weaver ID 
      # * +clazz+ a Ruby class name for a Weaver entity type
      # * +origin+ filename:lineno for error reporting
      def lookup_or_define(id, clazz, origin)
        val = child_entities[id]
        if val.nil?
          if clazz.kind_of?(Class)
            val = clazz.new(id, self)
            val.set_origin(origin, true)
          elsif clazz.kind_of?(Symbol)
            val = root_entity.prototype_instance(clazz, id, self, origin)
            raise WeaverError.new("Unknown type '#{clazz}") if val.nil?
            raise WeaverError.new("Prototype problem; asked for '#{clazz}' got '#{val.type}'") unless clazz == val.type
            raise WeaverError.new("Prototype problem; prototype of '#{clazz}' has validators") unless val.validators.size == 0 # TODO Remove
          else
            raise WeaverError.new("Unknown type '#{clazz}")
          end
        end
        
        return val
      end
      
      # Find the entity matching this ID.
      def scoped_lookup(id)
        # Check local child entities
        entity = child_entities[id]
        return entity unless entity.nil?
        
        # Check ourself.  We check ourselves second because an entity may contain an entity
        # with the same name as itself.
        if @id == id
          return self
        end
        
        # Check outer scope
        if @scope.nil?
          return nil
        end
        
        return scope.scoped_lookup(id)
      end

      # What are this entities siblings.  If +type+ is a Group this means other group members, otherwise other children of the same containment.
      # For non-Group types these don't need to be direct siblings with the same parent; if we ask for the siblings of a
      # Component with +type+ +:topology+ we will get all Components in the Topology      
      # * +type+ a Symbol such as +:infrastructure+
      def siblings(type)
        prototype = root_entity.prototype(type)
        raise WeaverError.new("Unknown Weaver extended type #{type}") if prototype.nil?
        return prototype.kind_of?(Group) ? group_siblings(type) : hosting_siblings(type)
      end
      
      # The siblings of the entity in containment.  This is not the simple siblings relationship -- having the same container --
      # but includes entities of different generations that are contained by the first instance of +type+ in the scope chain.
      # * +type+ a Symbol such as +:infrastructure+
      def hosting_siblings(type)
        retval = []
        ancestor = first_instance_of_weaver_type_in_scope(type)
        return retval if ancestor.nil? # We have no siblings within type if the type doesn't host us
        
        return ancestor.flatten_children.select { |sibling| !sibling.equal?(self) && sibling.type == self.type }
      end

      # Follows the scope chain back to the root_entity stopping at the first instance of type +type+.
      # * +type+ a Symbol such as +:infrastructure+
      def first_instance_of_weaver_type_in_scope(type)
        entity = self
        while !entity.nil? && !entity.weaver_instance_of?(type)
          entity = entity.scope
        end
        return entity
      end
      
      # The siblings of the entity within a group of type +type+.
      def group_siblings(type)
        retval = []
        
        # We only check direct containment and members
        @member_of.each do |id, container|
          if container.type == type
            container.members.values do |member|
              retval.push(member) unless self.equal?(member)
            end
          end
        end
        
        return retval
      end

      # Follow 'scope' until we encounter 'clazz', return nil if not found
      def scope_of_type(clazz)
        entity = scope
        while !entity.nil? && !entity.kind_of?(clazz)
          entity = entity.scope
        end
        return entity
      end
      
      # Returns all direct and indirect children of the entity.
      # RECURSIVE
      def flatten_children(accum = nil)
        accum = {} if accum.nil?
        child_entities.each_value do |child|
          if child.kind_of?(Base)
#            puts "Considering #{child.fully_qualified_id}"
            unless accum.member?(child.object_id) 
              accum[child.object_id] = child
              child.flatten_children(accum)
#            else
#              puts "Rejecting #{child.fully_qualified_id} because object_id #{child.object_id} already in #{accum.keys.inspect}"
            end
          end 
        end
        return accum.values
      end
      
      # Returns all attributes of this entity as well as the attributes of all direct and indirect contained child entities.
      def flatten_attributes
        retval = []
        amd.each do |id, att|
#          val = attributes[id]
          retval << att 
        end
        flatten_children.each do |child|
          child.amd.each do |id, att|
#            val = child.attributes[id]
            retval << att 
          end
        end
        return retval
      end
      
      # The 'fully qualified' id.  RECURSIVE
      def fully_qualified_id(pierce_imports = false)
        return (scope.nil? || scope.kind_of?(Root) || (!pierce_imports && scope.kind_of?(Import))) ? 
          id.to_s : scope.fully_qualified_id + '.' + id.to_s
      end
      
      # The id of a node, relative to a containing node.  RECURSIVE
      def relative_id(basis)
        return (scope.nil? || scope.kind_of?(Root) || scope.equal?(basis)) ? 
          id.to_s : scope.relative_id(basis) + '.' + id.to_s
      end

      # Is this a type or subtype of a Weaver registered type.  RECURSIVE
      # * +weaver_type+ a Symbol such as +:infrastructure+
      def weaver_instance_of?(weaver_type)
        return true if type == weaver_type
        return provenance.nil? ? false : provenance.weaver_instance_of?(weaver_type)
      end
      
      # Walk up the scope, looking for ImportContainers.  
      # We don't have a single place to hold all imports; Each Weaver file that does importing has its own
      # collection of imported files, so that possible collisions don't change the meanings of files
      # when they are imported by files having entities of the same Weaver ID.
      def import_by_symbol(sym)
        container = self
        while !container.nil?
          if container.is_a?(ImportContainerMixin) && !container.weaver_imports[nil].nil?
            unless container.weaver_imports[nil].nil?
              importable_entity = container.weaver_imports[nil].child_entities[sym]
              if importable_entity.kind_of?(Base)
                entity = Weaver::Transforms::Duplicate.duplicate(importable_entity)
                entity.provenance = importable_entity
                return entity
              end
            end
          end
          
          container = container.scope
        end
          
        return nil
      end

      # Set the +weaver_origin+ of entity.
      # +deeply+ if +true+, sets the origin of contained objects
      # RECURSIVE
      def set_origin(origin, deeply)
        @weaver_origin = origin
        
        # If this is the first time we have seen this symbol, make sure the attributes and child entities refer to nothing
        # so they inherit the current line, not their origin in some +type+ statement.
        if deeply
          # Set attributes.  Also set contained entities (for example, Node contains NetworkInterfaceController)
          child_entities.each do |_, child|
            child.set_origin(nil, true)
          end
        end
      end
      
      # Returns a Base or Attribute named by +path+
      # +path+ a (usually dotted) reference to a \Weaver Base or Attribute
      def resolve(path)
        match_data = /([^\.]*)/.match(path)
        return nil if match_data.nil?
        child = child_entities[match_data[1].symbolize]
        return nil if child.nil?
        return child if match_data.post_match.empty?
        return child.resolve(match_data.post_match[1..-1])
      end
      
      DEFAULT_TAG = :_default
      
      def tagged_with?(acceptable_tags)
        my_tags = @tags.empty? ? [ DEFAULT_TAG ] : @tags
        acceptable_tags = acceptable_tags.empty? ? [ DEFAULT_TAG ] : acceptable_tags
        # if there is an entry in atags    
        return !my_tags.select { |tag| acceptable_tags.include?(tag) }.empty?
      end
      
      # Get Provider of the appropriate type and (optional) name.  If ambiguous, return nil.
      # @param provider_class [Symbol] the Provider's provisioner_class, typically :artifact or :cloud
      # @param provider_id [Symbol] the Provider's id, or nil if we should use the (possibly unspecified) local Provider
      # @return [Provider|nil]
      def effective_provisioner(provider_class, provider_id = nil)
        # TODO deprecate provisioner() and provisioner=()
        
        # If no provider is specified, see if the provider_id is implicit because there is a local definition which acts as default
        if provider_id.nil? && self.is_a?(ProviderContainerMixin)
          candidates = providers.values.select { |provider| provider.provider_class == provider_class }
          provider_id = candidates[0].id if candidates.size == 1
        end
        
        inherited_provisioner = scope.nil? ? nil : scope.effective_provisioner(provider_class, provider_id)
        if inherited_provisioner.nil?
          return nil unless self.is_a?(ProviderContainerMixin)
          if provider_id.nil?
            candidates = providers.values.select { |provider| provider.provider_class == provider_class }
            return candidates[0] if candidates.size == 1
          else
            return providers[provider_id] if providers.key?(provider_id) && (provider_class.nil? || providers[provider_id].provider_class == provider_class)
          end
        elsif self.is_a?(ProviderContainerMixin) && providers.key?(provider_id)
          # We have a local one, process overrides
          prov = inherited_provisioner.clone
          providers[provider_id].amd.each do |attrid, attr|
            unless attr.implicit?
              dup = attr.clone
              dup.scope = prov
              prov.amd[attrid] = dup
            end 
          end
          return prov
        else
          return inherited_provisioner
        end
        
        return nil
      end

      # Provides access to the config[]
      def config
        if root_entity.nil? || root_entity.environment.nil? || root_entity.environment.project.nil?
#          raise WeaverError.new("Missing Environment; config[] unavailable")
          @@logger.debug("Missing Environment; config[] unavailable")
          return {}
        end
                
        return root_entity.environment.project.config_params
      end
      
      private
      
      def freshen_member_of
        @member_of.each do |id, group|
          if group.nil?
            group = root_entity.flatten_children.find { |entity| entity.id == id }
            member_of[id] = group unless group.nil?
          end
        end
      end
      
      # Remove kernel methods to get something like a Blank Slate pattern so that
      # Weaver users don't need to know about and avoid obscure Ruby ::Object methods.       
      [:test, :methods, :define_singleton_method, :untrust, :extend, :protected_methods,
      :tap, :singleton_class, :public_method, :untaint, :display, :singleton_methods,
      :taint, :instance_variables, :rand, :gets, :readlines, :exec, :select,
      :system, :lambda, :throw, :fail, :catch, :abort, :require_relative, :proc, :srand,
      :set_trace_func, :exit, :open, :spawn, :trace_var, :autoload, :trap, :format, :fork,
      :syscall, :load, :warn, :sleep].each do |meth|
        begin 
          undef_method meth
        rescue NameError
          # Do nothing, the list is for Ruby 1.9.2 and the actual implementation might not have
          # all the methods.  It isn't sufficient to check #respond_to? because Object doesn't
          # respond_to? :test even though it's there.
        end
      end
      
      
    end
    
  end
end