require 'weaver/core/base.rb'
require 'weaver/core/depends_on_mixin'

module Weaver
  module Core
    # Automation represents an automation (shell script, Chef recipe, etc)
    #
    # Author:: Ed Snible
    # Author:: Florian Rosenberg
    class Automation < Base
      include DependsOnMixin
      include Generation

      # the name of the executable
      attr_accessor :executable
       
      # type of script :chef, :bash, :sh, ...
      attr_accessor :script_type  
         
      # style for parameter passing, :long_style, :short_style, etc
      attr_accessor :parameter_style
      
      # Credential; defaults to root 
      attr_accessor :run_as 
      
      # * +id+ Symbol to represent the entity
      # * +parent+ The enclosing Base
      def initialize(id, parent)
        raise WeaverError.new("Automation id problem, id is a #{id.class}") unless id.kind_of?(Symbol) || id.kind_of?(String)
        super(id, parent)
        
        @type = :automation
        @executable = id
        @parameter_style = nil
        @run_as = 'root'
        
        @prefix_comments = []
        @inner_comments = []
      end

      # Returns true if this Automtion is equal to another Automation
      def ==(another_script)
        if !another_script.kind_of?(Automation)
          return false
        end
        retval = super
        retval = retval && @executable == another_script.executable && @type == another_script.type && 
#        @unresolved_parameters == another_script.unresolved_parameters &&
          @parameter_style == another_script.parameter_style && @run_as == another_script.run_as
        return retval
      end
      
      # Describe this Automation
      def to_s
        return "Automation :#{@executable}"
      end
      
      # Validates the entity and its direct and indirect contained children.
      # * +out_messages+ ProblemStatus will be pushed onto this for any validation message
      # * +excluded_phases+ optional Array which may be +[:predefined]:
      def validate_collect(out_messages, desired_phases = nil)
        @@logger.debug "Validating script '#{executable}'"
        @@logger.debug "  has type '#{@type}'" #unless type?
        @@logger.debug "  has parameter_style '#{@parameter_style}'" #unless parameter_style?
        @@logger.debug "  has run_as '#{@run_as}'" #unless run_as?
        
        super
      end

      # The node this script runs upon
      def node
        return scope.scope_of_type(Node)
      end
      
      # This lets the old "parameters['jke.db_hostname']" syntax still work with Attributes
      def parameters
        return ParametersEvaluator.new(self)
      end
      
      # @private
      # This lets the old "parameters['jke.db_hostname']" syntax still work with Attributes
      # TODO Deprecate
      class ParametersEvaluator # :nodoc:
        def initialize(entity)
          @entity = entity
        end
        
        def [](attrib_name)
          attr = @entity.amd.values.find { |obj| obj.name == attrib_name }
          return attr.nil? ? nil : attr.value
        end
      end
    end
  end
end