require 'weaver/core/base'
require 'weaver/core/generation'
require 'weaver/core/automation_container'
require 'weaver/core/component_container_mixin'
require 'weaver/core/policy_container_mixin'
require 'weaver/core/import_container_mixin'
require 'weaver/core/topology_container_mixin'
require 'weaver/core/provider_container_mixin'
require 'weaver/core/node_container_mixin'
require 'weaver/core/user_container_mixin'
require 'weaver/core/group_container_mixin'
require 'weaver/core/link_container_mixin'
require 'weaver/core/service_container_mixin'
require 'weaver/core/artifact_container_mixin'
require 'weaver/core/storage_container_mixin'

module Weaver
  module Core
    # Represents a topology, such as an Infrastructure or Application
    #
    # @see Weaver::Dsl::TopologyStatement
    #
    # Author:: Ed Snible
    class Topology < Base
      include Weaver::Core::Generation
      include ServiceContainerMixin
      include StorageContainerMixin
      include ProviderContainerMixin
      include GroupContainerMixin
      include UserContainerMixin
      include TopologyContainerMixin
      include NodeContainerMixin
      include PolicyContainerMixin
      include LinkContainerMixin
      include ArtifactContainerMixin
      
      # After cloning Topology.startup_order is stale, this method updates it
      register_reference_freshener :freshen_startup_order
      
      # a Weaver::Core::Cloud
      attr_accessor :cloud
      
      # Startup order of the nodes in the application
      attr_accessor :startup_order
      
      # Array of length=1 Hash of source-path to symbol for renamings
      # In the future we may refactor this into Base but currently we only need this to generate
      # renamings in the Infrastructure Erubis generator
      attr_reader :unresolved_renamings
      
      # Hash of source to target, as Strings
      attr_reader :unresolved_realizations
      
      # Used to augment attributes
      attr_reader :unresolved_additional_declarations
      
      # Used to hold raw responses from deployment system
      attr_accessor :cloud_status_internal
      
      # * +id+ A symbol for the Topology, e.g. +:node1+
      # * +parent+ The enclosing Base
      def initialize(id, parent)
        super(id, parent)
        @type = :topology
        @unresolved_renamings = []
        @unresolved_realizations = {}
        @startup_order = []
        @unresolved_additional_declarations = []
        @provisioner = nil
        @cloud_status_internal = {}
        
        amd[:version] = Attribute.new(:version, self, true)
        amd[:version].implicit = true
        amd[:version].required = false
      end

      # All Weaver entities define an +initialize_copy+ to deeply copy their children.
      def initialize_copy(other)
        super
        
        @unresolved_additional_declarations = unresolved_additional_declarations.clone
        @unresolved_realizations = unresolved_realizations.clone
        @unresolved_renamings = unresolved_renamings.clone
      end

      # Validate entity      
      # * +out_messages+ ProblemStatus will be pushed onto this for any validation message
      # * +desired_phases+ optional Array which may be +[:predefined]:
      def validate_collect(out_messages, desired_phases = nil)
        @@logger.debug "Validating #{type} #{id.inspect}"

        super
        
        # We must validate policies explicitly because PolicyContainerMixin does not register a handler
        validate_policies(out_messages, desired_phases);
        
        if topologies.empty? && nodes.empty? && scope.kind_of?(Root)
          out_messages.push Weaver::Validation::ProblemStatus.new(Logger::WARN, "The topology cannot be deployed because it has no nodes", self)
        end
        
        @startup_order.flatten.each do |node|
          if node.instance_of? Symbol
            out_messages.push Weaver::Validation::ProblemStatus.new(Logger::ERROR, "Unknown Node in startup_order: #{node}", self, startup_order.instance_variable_get(:@weaver_origin))
            return
          elsif node.nil?
            out_messages.push Weaver::Validation::ProblemStatus.new(Logger::ERROR, "'nil' in startup_order", self, startup_order.instance_variable_get(:@weaver_origin))
          end
        end
        
        if @startup_order.flatten.size != @startup_order.flatten.uniq.size
          out_messages.push Weaver::Validation::ProblemStatus.new(Logger::ERROR, "Duplicate node in startup_order", self, startup_order.instance_variable_get(:@weaver_origin))
#        elsif !infrastructure.nil? && startup_order.flatten(1).size < infrastructure.unique_nodes.size
#          out_messages.push Weaver::Validation::ProblemStatus.new(Logger::ERROR, "Infrastructure has #{infrastructure.unique_nodes.size} nodes, startup_order has only #{startup_order.size}", 
#          self, startup_order.instance_variable_get(:@weaver_origin))
        else
          # Now analyze data flow to see if script parameter dependencies can be satisfied if
          # node[x].ip_address and node[x].hostname change in the startup_order
          unless startup_order.empty? || Weaver::Service::PropagationChecker.valid_startup_order(self)
            candidate = Weaver::Service::PropagationChecker.generate_startup_order(self)
            if candidate.nil?
              out_messages.push Weaver::Validation::ProblemStatus.new(Logger::ERROR, "Scripts depend on properties that cannot be satisfied starting in startup_order; there is no valid startup_order for this environment", 
              self, startup_order.instance_variable_get(:@weaver_origin))
            else
              out_messages.push Weaver::Validation::ProblemStatus.new(Logger::ERROR, "Scripts depend on properties that cannot be satisfied starting in startup_order; a valid order is startup_order = #{candidate.inject("") {|str, id| str + ', :' + id.to_s }[2..-1]}", 
                self, startup_order.instance_variable_get(:@weaver_origin))
            end
          end
        end
        
      end
      
      # Returns true if this Topology is equal to another Topology
      def ==(other)
        retval = super
        return retval && nodes == other.nodes && 
          links == other.links && users == other.users && cloud == other.cloud
        # TODO Groups, Topologies, Policies
      end

      # Return the controlling Provider.  This may be inherited.  If there are multiple ones
      # the provisioner is ambiguous but we will return one anyway, leaving any detection
      # of the problem to validation.
      def provisioner
        return @provisioner if @provisioner                     # explicit reference

        return providers.values.to_a[0] unless providers.empty? # locally defined provisioner
        
        # If there are any nodes with provisioner, the topology is considered an instance of the node provisioner
        node_with_provisioner = nodes.values.find { |node| node.scope == self && node.provisioner }
        return node_with_provisioner.provisioner unless node_with_provisioner.nil?

        if scope.respond_to?(:provisioner)
           return scope.provisioner unless scope.provisioner.nil? # scope's provisioner
        end
        
        return nil
      end
      
      # Explicitly reference a provisioner instead of inheriting one from parent
      # - +prov_ref+ Core::Provider
      def provisioner=(prov_ref)
        @provisioner = prov_ref
      end
      
      # Called as part of Duplicate.duplicate to ensure Topology.startup_order refers to the duplicate, not original, nodes      
      def freshen_startup_order(old_to_new)
        @startup_order = @startup_order.map { |ele| ele.respond_to?(:map) ? ele.map { |node| old_to_new.key?(node) ? old_to_new[node] : node } : old_to_new.key?(ele) ? old_to_new[ele] : ele }
      end
      
      # Return Node
      def node_by_id(id)
        raise WeaverError.new("'id' must by a Symbol but it is a #{id.class}") unless id.kind_of?(Symbol)
        retval = flatten_children.find { |n| n.kind_of?(Node) && n.id == id }
        return retval.nil? ? id : retval
      end

      # Summarize
      def to_s
        "#{type}[id: #{id}, name: #{name}, description: #{description}]"
      end
      
      # Return Array of Topology
      def get_cloud_infrastructures()
        infs = []
        infs.push(self) unless self.cloud.nil? && self.provisioner.nil?
        flatten_children.each { |entity| infs.push(entity) if entity.kind_of?(Topology) && (entity.cloud != nil || entity.provisioner != nil) }
        return infs
      end
      
    end
  end
end