require 'weaver/core/logging'
require 'weaver/core/node'
require 'weaver/dsl/node_statement'

module Weaver
  module Dsl
    # Mix in NodeContainer in statements that may contain {NodeStatement}
    #
    # Author:: Ed Snible
    # Copyright:: IBM
    module NodeContainer
      include Weaver::Core
      
      # * *Args* :
      #   -+base+ the AbstractStatement being mixed with
      def self.included(base)
        base.register_statement_handler([:node_handler])
      end

      # * *Args* :
      #   -+sym_or_model+ <Symbol representing Weaver type, e.g. <code>:node</code> | Core::Node
      #   -+id+ Symbol naming entity
      #   -+block+ optional block of additional metadata
      def node_handler(sym_or_model, id, &block)
        return nil unless is_a?(Weaver::Dsl::NodeContainer) && (sym_or_model.kind_of?(Node) || root_statement.model.weaver_extended_type?(sym_or_model, Node))
        return nil unless id.kind_of?(Symbol) 
        
        prototype = sym_or_model.kind_of?(Symbol) ? root_statement.model.prototype_instance(sym_or_model, id, model, weaver_stackframe) : sym_or_model 
        node(id, prototype, &block)
      end

      # If prototype is nil, we will create a core Node, otherwise inherit attributes,
      # default attribute values, and validators from the prototype
      def node(id, prototype = nil, &block)
        @@logger.debug "node[(#{prototype.nil? ? 'node' : prototype.type}): #{id}]"
        
        return if fail_id_redefinition(id, Node)
        
        node_model = @model.nodes[id]
        if node_model.nil?
          node_model = prototype || root_statement.model.prototype_instance(:node, id, model, weaver_stackframe)
          if @model.respond_to?(:provisioner) && !@model.provisioner.nil?
            if node_model.provider_mixins.empty?
              @model.provisioner.mixin_provider_attributes(node_model)
            end
          end
          @model.nodes[id] = node_model
        else
          root_statement.model.widen(node_model, prototype) unless prototype.nil?
        end
        
        unless block.nil?
          node = NodeStatement.new(node_model, self)
          
          begin
            node.instance_eval &block
          rescue ArgumentError => ae
            # For example the statement 'component' without an ID yields
            # 'wrong number of arguments (0 for 1)'
            node_model.add_validation_failure(" #{ae.message}", weaver_stackframe(ae.backtrace))
            @@logger.warn "Caught #{ae.inspect} processing 'node' statement, backtrace=#{ae.backtrace}"
          rescue NameError => ne
            # For example the statement 'N1.name' yields NameError uninitialized constant
            node_model.add_validation_failure(" #{ne.message}", weaver_stackframe(ne.backtrace))
            @@logger.warn "Caught #{ne.inspect} processing 'node' statement, backtrace=#{ne.backtrace}"
          end
        end

        return node_model
      end
      
    end

  end
end
