require 'weaver/dsl/component_statement'
require 'weaver/dsl/depends_on_statement'
require 'weaver/dsl/link_statement'
require 'weaver/dsl/firewall_statement'
require 'weaver/dsl/nic_statement'
require 'weaver/dsl/automation_container'
require 'weaver/dsl/provider_container_statement'
require 'weaver/dsl/service_container_statement'

module Weaver
  module Dsl
    # The NodeStatement implements the processing of the language element 'node'.
    # Nodes support statements: include, component, os, firewall, cloud, realized_by
    #
    # Author:: Ed Snible
    # Author:: Florian Rosenberg
    # @see Node
    class NodeStatement < AbstractStatement
      include ServiceContainerStatement
      include ProviderContainerStatement
      include PolicyContainer
      include AutomationContainer
      include ComponentContainer
      include DependsOnStatement

      # * +id+ Symbol representing the entity, or a Core::Node which is the entity
      # * +parent_statement+ The enclosing AbstractStatement
      def initialize(id, parent_statement)
        @@logger.debug "node[id: #{id} on #{parent_statement.model.type} :#{parent_statement.model.id}]"
        super(id, :node, parent_statement)
      end
      
      # Parse "realized_by <symbol|nodepath>", symbol refers to another Node typically an Infrastructure Node
      def realized_by(reference)
        if reference.kind_of?(NodeStatement)
          @@logger.debug "realized_by[#{reference.model.type} (#{reference.model.id.inspect}) {...}]"
          @@logger.info "'realized_by' is deprecated; replace with 'realizes' in enclosing topology"
          model.realized_by << reference.model
        elsif reference.kind_of?(Symbol)
          @@logger.debug "realized_by[#{reference}]"
          @@logger.info "'realized_by' is deprecated; replace with 'realizes' in enclosing topology"
          @model.realized_by << @model.root_entity.node_by_id(reference)
        else
          # FailureEaters are reported elsewhere and shouldn't get an addition error here
          model.add_validation_failure("realized_by needs a symbol or node but a #{reference.class} was supplied") unless reference.kind_of?(FailureEater)
        end
      end
      
      # Parse cloud :rc2, :url => 'https://rc2-v3.cloud9.ibm.com/rc2api/rest'
      def cloud(type, args = nil)
        @@logger.debug "[cloud type: #{type.inspect} args: #{args.inspect}]"        
        args = type if args == nil

        # inherit type from parent if nil
        if type.kind_of?(Hash)          
         cloud = Cloud.new(@model.scope.cloud.nil? ? nil : @model.scope.cloud.type) 
         cloud.url = @model.scope.cloud.url unless @model.scope.cloud.nil?
         cloud.description = @model.scope.description
        else
          cloud = Cloud.new(type)
          cloud.url = args[:url]
          cloud.description = args[:description]
        end
        cloud.image = args[:image]
        @model.cloud = cloud
        
        if model.amd[:image].nil?
          model.amd[:image] = Attribute.new(:image, model)
        end
        model.image = args[:image]
        model.amd[:image].implicit = true
      end

      # Parse os :linux, :flavor => :rhel, :version => '6.1', :architecture => :x86_64
      def os(os_type, os_data)
        @@logger.debug "os type: #{os_type} data: #{os_data.inspect}]"
        os = OperatingSystem.new
        os.type = os_type
        os_data.each { |key, value|
          writemethod = key.to_s + "="
          if os.class.method_defined?(writemethod) then
            method = os.method(writemethod)
            method.call(value)
          else
            raise WeaverError.new("Unknown 'os' statement property: ':#{key}'.")
          end
        }
        model.os = os
      end

      # Parse firewall { } using a FirewallStatement
      def firewall(&block)
        return model.add_validation_failure("Firewall redefinition") unless @model.firewall.nil?
                
        firewall_statement = FirewallStatement.new(self)
        @model.firewall = firewall_statement.model
        firewall_statement.instance_eval &block unless block.nil?
      end

      # Parse requires { }      
      def requires(requires_data)
        @@logger.debug "requires [#{requires_data}]"               
        handler = Weaver::Dsl::Requires::TypeHandlerRegistry.instance.find(requires_data.keys[0])        
        unless handler 
          return model.add_validation_failure("Cannot find an appropriate 'requires' handler.")
        end
        
        # expect to have one handler
        requirement = handler.handle(self, requires_data)
        requirement.weaver_origin = weaver_stackframe
        
        return model.add_validation_failure("Can't create requirement from #{requires_data}") unless requirement.kind_of?(Requirement)

        @model.requirements.push(requirement)
      end

      # Parse nic (:<ID>) { }
      # * +id+ A symbol for the NetworkInterfaceController, e.g. +:nic0+
      # * +block+ A block containing Weaver statements about the NIC
      def nic(id, &block)
        return model.add_validation_failure("Syntax: nic (:id) {}") unless id.kind_of?(Symbol)
        return if fail_id_redefinition(id, NetworkInterfaceController)
        
        nic_model = model.nics[id]
        if nic_model.nil?
          nic_model = root_statement.model.prototype_instance(:nic, id, model, weaver_stackframe)
          model.nics[id] = nic_model
        end
        
        unless block.nil?
          nic = NicStatement.new(nic_model, self)
          
          begin
            nic.instance_eval &block
          rescue ArgumentError => ae
            # For example the statement 'automation' without an ID yields
            # 'wrong number of arguments (0 for 1)'
            nic_model.add_validation_failure(" #{ae.message}", weaver_stackframe(ae.backtrace))
            @@logger.warn "Caught #{ae.inspect} processing 'nic' statement, backtrace=#{ae.backtrace}"
          end
        end

        return nic_model
      end
      
      # Summarize
      def to_s
        "node[id: #{@id}, name: #{@name}, description: #{@description}]"
      end
      
    end

  end

end
