require 'erubis'
require 'string'
require 'json'
require 'devops_deployment_lib/providers/iwd_virtual_system_client'
require "weaver/core/logging"
require 'weaver/core/attribute'
require 'weaver/core/component'
require 'weaver/core/automation'
require 'weaver/core/operating_system'
require 'weaver/core/os/operating_system_handler'
require 'weaver/core/weaver_error'
require 'weaver/core/late_binding'
require 'weaver/chef/chef_automation_handler'
require 'weaver/iwd/pattern_utils'

module Weaver
  # \Weaver::Iwd provides classes for working with {IBM Workload Deployer}[http://www-01.ibm.com/software/webservers/workload-deployer/]
  #
  # @see IwdPatternUpdateComparator
  # @see IwdRequestGenerator
  # @see IwdWeaverInfGenerator
  # @see IwdWebmockGenerator
  module Iwd
    # Generates a IWD pattern instantiation JSON request from a Weaver
    # infrastructure definition.
    #
    # Author:: Florian Rosenberg
    #
    class IwdRequestGenerator
      include Weaver::Core
      
      # Creates a new IwdRequestGenerator for a given Environment definition.
      #
      # * *Args*    :
      #   - +environment_folded+ -> A Topology with Weaver type +:cloud_pattern+
      #
      # * *Returns* :
      #   - an IwdRequestGenerator instance
      #
      def initialize(environment_folded, iwd_version = nil)
        @pattern = environment_folded
        
        if @pattern.amd[:pattern_id].nil?
          raise WeaverError.new("No :pattern_id on #{@pattern}; cannot deploy to IWD")
        end  
       
        # these attributes get special treatment - they will not used in the JSON request attributes.
        @ignored_attributes = [ :name, :description, :internal_id, :part_name, :part_id, :image_id, :image, :pattern_id, 
          :cloud_group, :ip_group, :environment_profile, :ip_address, :hostname, :archive, :id_in_pattern, :script_id, 
          :multiplicity, :multiplicity_index, :vm_status, :config_status, :stack_state, :resource_state, :flavorid, :extra_ip_groups ]

        @pattern.unique_nodes.each do |n_id, n|
          if iwd_version.nil?
            Weaver::Iwd::PatternUtils::add_iwd_macros_to_node(@pattern.provisioner.version || @pattern.cloud.version, n)
          else
            @@logger.debug("Taking into account IWD version retrieved from server: #{iwd_version}")
            Weaver::Iwd::PatternUtils::add_iwd_macros_to_node(iwd_version, n)
          end
        end
      end

      # Generates a JSON request for an infrastructure definition.
      #
      # * *Returns* :
      #   - the JSON payload to send to IWD for instantiating a pattern
      def to_json()
        json = {}
        json['pattern'] = "/resources/patterns/#{@pattern.amd[:pattern_id].value}"
        json['name'] = @pattern.amd[:name].value
        add_parts(json)
        add_inf_attributes(json)
        #puts JSON.pretty_generate(json)
        return json.to_json()
      end

      private
      
      def add_parts(json)
        json["parts"] = []

        @pattern.unique_nodes.each do |k, v|
          part = {}
          @@logger.warn "#{v.fully_qualified_id} is not a cloud_node" if v.amd[:part_id].nil? || v.amd[:part_name].nil?  
          part['id'] = v.part_id.to_i unless v.amd[:part_id].nil?
          part['label'] = v.part_name unless v.amd[:part_name].nil?
          part['description'] = v.description()
          # IWD does not let us specify a count/memberCount for parts.  We may only create instances with the default number
          # specified in the pattern.
          part['virtualimage'] = "/resources/virtualImages/#{v.image}" unless v.image.nil?

          add_environment_profile_attributes(v, part)
          handle_automation(@pattern, v, part)          
          add_script_packages(v, part)
          add_part_properties(v, part)
          
          json["parts"] << part
        end
      end
      
      def handle_automation(pattern, node, part_json)
        handler = Weaver::Chef::ChefAutomationHandler.new()
        # convert each chef_node_attr that is a Hash to a String
        handler.interpolate_and_convert_chef_attributes(pattern, node)
        # If there are chef_automations, collect the attribute values to be used together in one chef automation
        # otherwise, attrs will hold default values
        attrs = handler.generate_chef_attributes(node)
        # Find DevOps bootstrap script package identified by the following properties
        #  "CHEF_NODE", "DEPS_FILE_URL", "CHEF_NODE_ATTR"
        #
        node.automation.select { |k,v|
          v.type == :script_package && v.amd[:chef_node] &&
          v.amd[:chef_node_attr] && v.amd[:deps_file_url] }.each do |k, v|
          # If there is chef automation, it will override chef node values that were specified in the infrastructure
          # The common case for specifying automation directly in the infrastructure is when a user
          # sets the infrastructure properties in an external yaml file
          if handler.has_chef_automation(node)
            # assign values
            # This may override a value set directly on a chef_node_attr script package automation in the infrastructure
            v.amd[:chef_node].value = attrs[:chef_node]
            v.amd[:chef_node_attr].value = attrs[:chef_node_attr]
            v.amd[:deps_file_url].value = attrs[:deps_file_url]
          else
            # if no chef automation is specified, set the chef script package parameters to default values ONLY if they are nil
              v.amd[:chef_node].value = v.amd[:chef_node].value || attrs[:chef_node]
              v.amd[:chef_node_attr].value = v.amd[:chef_node_attr].value ||  attrs[:chef_node_attr]
              v.amd[:deps_file_url].value = v.amd[:deps_file_url].value || attrs[:deps_file_url]
          end
        end
      end      
      
      def add_environment_profile_attributes(node, part_json)
        # determine if environmentProfile is used
        env_profile = @pattern.amd[:environment_profile]
        @@logger.warn "#{@pattern.fully_qualified_id} is not an IWD Pattern" if env_profile.nil?
        if env_profile && env_profile.value
          ip_group = node.amd[:ip_group]
          cloud_group = node.amd[:cloud_group]
          part_json[cloud_group.name] = "/resources/clouds/#{cloud_group.value}"
                    
          # This is a BUG in the documentation the ipGroup needs to be in a "nics" element like this                    
          # "nics": [
          #           [
          #             {
          #               "ipGroup": "/resources/ipGroups/2"
          #             }
          #           ]
          #         ],
          nics_json = [ [ {  ip_group.name => "/resources/ipGroups/#{ip_group.value}"} ] ] 
          if node.amd[:extra_ip_groups]
            if node.amd[:extra_ip_groups].value.respond_to?(:map)
              nics_json[0].concat(node.amd[:extra_ip_groups].value.map { |grp| { ip_group.name => "/resources/ipGroups/#{grp}"} })
            elsif node.amd[:extra_ip_groups].value != nil
              nics_json[0].push( { ip_group.name => "/resources/ipGroups/#{node.amd[:extra_ip_groups].value}"} )
            end
          end
#          if node.amd[:extra_ip_groups] && node.amd[:extra_ip_groups].value.respond_to?(:map)
#            nics_json[0].concat(node.amd[:extra_ip_groups].value.map { |grp| {  ip_group.name => "/resources/ipGroups/#{grp}"} })
#          end
          part_json["nics"] = nics_json
            
          unless node.amd[:flavorid].nil?
            part_json['flavorid'] = node.flavorid
          end
        end
      end

      def add_script_packages(node, part_json)
        part_json['scripts'] = []

        node.automation.select { |k, v| v.type == :script_package && v.amd.length > 0 }.each do |k, v|
          s_json = {}
          s_json['id'] = v.id_in_pattern.to_i
          s_json['label'] = v.name
          s_json['description'] = v.description
          add_script_parameters(v, s_json)
          part_json['scripts'] << s_json
        end

      end

      def add_script_parameters(script_pkg, json)
        json['parameters'] = []
        script_pkg.amd.each do |k, v|
          unless @ignored_attributes.include?(v.id) || v.name.nil?
            p_json = {}
            newval = v.value || v.default || ""
            newval = newval.respond_to?(:ciphertext) ? Weaver::Utils::ValueDecryptor.decrypt(newval, @keyfile) : newval
            p_json['value'] = newval.to_s
            p_json['key'] = v.name
            json['parameters'] << p_json
          end
        end
      end

      def add_part_properties(node, part_json)
        part_json['properties'] = []
        node.amd.each do |k, v|
          unless @ignored_attributes.include?(v.id)
            p_json = {}
            p_json['key'] = v.name
            p_json['label'] = v.label || v.name
            p_json['pclass'] = v.group
            p_json['description'] = v.description
            
            # process value and decrypt if necessary - each value needs to be serialized as string even if its boolean or integer (IWD problem)
            if v.value != nil
              newval = v.value.respond_to?(:ciphertext) ? Weaver::Utils::ValueDecryptor.decrypt(v.value, @keyfile) : v.value
              p_json['value'] = newval.to_s
            elsif v.default != nil
              newval = v.default.respond_to?(:ciphertext) ? Weaver::Utils::ValueDecryptor.decrypt(v.default, @keyfile) : v.default
              p_json['value'] = newval.to_s
            else
              newval = ""
              p_json['value'] = newval
            end
            
            # set value type in JSON based on Ruby type (use newval to be able to determine the data type of 
            # the decrypted value)
            #puts "Value for #{v.name}: #{newval} (#{newval.type})"
            p_json['type'] = convert_attr_type(newval)
           
            ## determine valid values
            valid_values = v.validators.find { |entry| entry[0] == :valid_values } # finds the whole validator array
            p_json['validValues'] = valid_values.slice(1..-1).map { |v| v.to_s } if valid_values # make it a string, IWD complains if not.
            valid_values = nil
            part_json['properties'] << p_json              
          end
        end
      end

      def convert_attr_type(value)
        return "integer" if value.kind_of?(Fixnum)
        return "string" if value.kind_of?(String)
        return "boolean" if value.kind_of?(TrueClass) ||  value.kind_of?(FalseClass)
      end

      def add_inf_attributes(json)
        cloud_group = @pattern.amd[:cloud_group]
        env_profile = @pattern.amd[:environment_profile]
                
        # check which one is used
        if cloud_group && cloud_group.value
          json[cloud_group.name] = "/resources/clouds/#{cloud_group.value}"
        elsif env_profile && env_profile.value 
          json[env_profile.name] = "/resources/environmentProfiles/#{env_profile.value}"
        elsif cloud_group && cloud_group.default
          json[cloud_group.name] = "/resources/clouds/#{cloud_group.default}"
        elsif env_profile && env_profile.default
          json[env_profile.name] = "/resources/environmentProfiles/#{env_profile.default}"
        end        
      end

    end
  end
end
