require 'erubis'
require 'string'
require 'date'
require 'pathname'
require 'devops_deployment_lib/providers/iwd_virtual_system_client'
require 'devops_deployment_lib/http_error'

require 'weaver/core/attribute'
require 'weaver/core/component'
require 'weaver/core/automation'
require 'weaver/core/operating_system'
require 'weaver/core/os/operating_system_handler'
require 'weaver/core/weaver_error'
require 'weaver/iwd/iwd_request_generator'
require 'weaver/iwd/pattern_utils'
require 'weaver/core/logging'
require 'weaver/zephyr/weaver_generator'
require "rexml/document"
require "weaver/transforms/duplicate"
require "weaver/core/weaver_transforms"
require 'weaver/utils/yaml_property_writer'

module Weaver
  # Contains all IWD specific code
  module Iwd
    # Implements an adapter to IWD that allows to create a Weaver infrastructure
    # definition from an IWD pattern. The pattern concepts are mapped to the
    # Weaver infrastructure concepts as outlined below.
    #
    # ===Mapping IWD Pattern to Weaver Infrastructure Definition
    #
    # TODO describe mapping in more detail
    #
    #
    # Author:: Florian Rosenberg
    #
    class IwdWeaverInfGenerator
      include Weaver::Core

      # IWD hostname or fully-qualified hostname
      attr_reader :host
      
      # IWD user name
      attr_reader :username
      
      # IWD password
      attr_reader :password
      
      # If true, a valid SSL certificate will be required
      attr_reader :verify_ssl_cert
      
      # Creates a new IwdWeaverInfGenerator for a given IWD server that allows
      # to create Weaver infrastructure definition files.
      #
      # * *Args*    :
      #   - +host+ -> the IWD server hostname (String)
      #   - +username+ -> the user for the server (String)
      #   - +password+ -> the password for the server (String)
      #   - +verify_ssl_cert+ -> flag to indicate whether to validate SSL
      #                          certificates of the server (Boolean; default: false)
      # * *Returns* :
      #   - A newly created IwdWeaverInfGenerator
      #
      def initialize(host, username, password, verify_ssl_cert = false, &reporter)       
        @host = host
        @username = username
        @password = password
        @verify_ssl_cert = verify_ssl_cert
        @reporter = reporter || Proc.new { |lvl, status, msg| }
        @client = DevopsDeploymentLib::IwdVirtualSystemClient.new(host, username, password, { :log_level => Weaver::Core::Logging.get_level })
        if !verify_ssl_cert
          @client.ssl_verify_mode=(OpenSSL::SSL::VERIFY_NONE)
        end          
        flush_icon_topologies_cache()
        @env_profile_readable = true # will be checked later - assume true for now
      end

      # Generates an infrastructure topology for an IWD Virtual System Pattern. The 
      # pattern may optionally also contain ICON topology information (default is true).
      #
      # * *Args*    :
      #   - +iwd_pattern_id+ -> the Virtual System Pattern id on the IWD server (Fixnum)
      #   - +use_image_topologies+ -> whether ICON topology information for the images should be used for the generation (Boolean)
      #
      # * *Returns* :
      #   - A newly created Infrastructure topology
      #
      def generate_infrastructure(iwd_pattern_id, use_image_topologies = true)
        begin 
          @@logger.debug("Determining server variant and version...")
          determine_server_variant_and_version()
  
          @@logger.debug("Generating infrastructure for pattern id: #{iwd_pattern_id}")
          json = @client.describe_pattern(iwd_pattern_id)
          raise WeaverError.new("Aborted") if @reporter.call(Logger::INFO, "Generating", "Weaver topology for pattern id: #{iwd_pattern_id}") == :abort
          #@@logger.debug("--start pattern_json------------")
          #@@logger.debug(json.to_json)
          #@@logger.debug("--end pattern_json------------")
          # get the root statement and parse the built-in types and being able to create them
          @root = Weaver::Dsl::WeaverParser.new().parse(nil)
          
          # Create the infrastructure object of type iwd_pattern
          inf = @root.prototype_instance(:cloud_pattern, json['name'].to_s.symbolize, nil, nil)
          inf.name = json['name']
          inf.description = json['description'] || ""
          inf.pattern_id = json['id']
          inf.prefix_comments << ""
          inf.prefix_comments << "Generated automatically by weaver on #{DateTime.now.strftime("%FT%R")}" 
          inf.prefix_comments << "from #{get_hostname()}/dashboard/design/patterns/##{iwd_pattern_id}"                                
          inf.prefix_comments << ""
          inf.prefix_comments << "DO NOT MANUALLY CHANGE THIS FILE.\n #"
              
          # Set the cloud attribute
          inf.cloud = Cloud.new(:iwd_vsp)
          inf.cloud.url = host
          inf.cloud.description = "IBM Workload Deployer/SmartCloud Provisioning"
          inf.cloud.version = @server_version
          
          # Set the Provider attribute
          iwd_provider = Provider.new(:iwd_vsp, inf)
          iwd_provider.host = host
          iwd_provider.version = @server_version
          inf.providers[:iwd_vsp] = iwd_provider 
          
          # Set the cloud group or environment profile option
          add_cloud_group_and_env_profile_attr(inf)
          
          startup_order = []
  
          # Map each part in JSON to an iwd_node in Weaver
          parts = json['parts'] || []
          parts.each do |part_json|
            part_id =  part_json['id']
            @@logger.debug("Processing IWD part id: #{part_id}")
  
            # get more detailed info for the part. the pattern json does not contain everything needed.
            raise WeaverError.new("Aborted") if @reporter.call(Logger::INFO, "Processing", "Part id: #{part_id}") == :abort
            part_json_details = @client.describe_pattern_part(iwd_pattern_id, part_id)
            
            node = @root.prototype_instance(:cloud_node, "#{part_json['label']}".symbolize, inf, nil)
            node.name = part_json_details['key']
            node.description = part_json['description']
            node.part_id = part_id.to_i
            node.image_id = part_json_details['resid'].to_s
            node.part_name = part_json_details['key']
            node.multiplicity = part_json_details['memberCount'] unless part_json_details['memberCount'].nil?
            if node.amd[:multiplicity] 
              elastic = part_json_details['elastic'] || 'nil'
              elasticDisabled = part_json_details['elasticDisabled'] || 'nil'
              node.amd[:multiplicity].prefix_comments.push("IWD elastic=#{elastic} / elasticDisabled=#{elasticDisabled}")
            end
                          
            # Part 'properties' as Weaver attributes
            part_json_details['properties'].each { |prop_json| add_attributes(node, prop_json) }
    
            add_cloud_group_and_ip_group(node)
  
            # Part OS information
            virt_img_location = part_json['virtualimage']  # get image URL and get
            # id: "virtualimage"=>"/resources/virtualImages/28"
            if virt_img_location
              virt_image_id = virt_img_location.slice(virt_img_location.rindex('/') + 1,  virt_img_location.length)
            # # remove the /resouces/.../ part
            else
              virt_image_id = :unknown
            end
  
            # Add cloud image info
            node.cloud = Cloud.new(:iwd_vsp)
            node.cloud.url = host
            node.cloud.image = virt_image_id
  
            unless virt_image_id == :unknown
              add_os_information(node, virt_image_id)
              if use_image_topologies
                add_icon_metadata(node, virt_image_id, part_json_details['resid'] )
              end
            end
  
            # Map each IWD script package to a script_package (subtype of
            # automation) type in Weaver
            add_script_packages(node, iwd_pattern_id, part_id)
  
            # add node to Infrastructure (and handle equal node names by appending
            # an integer as long as we find on that is available)
            i = 0
            while inf.nodes.key?(node.id)
              node.id = "#{node.id}_#{i}".symbolize
              i += 1
            end
            inf.nodes[node.id] = node
  
            process_node_startup_order(node, part_json_details, startup_order)
          end       
          
          WeaverTransforms::rubyize_ids(inf)    # In case the pattern included non-ASCII characters, for example
          
          # compact to Remove nils; for example pattern 167 on FIT2 had startup_order [nil, core_os_1].
          inf.startup_order = startup_order.compact
          return inf        
        rescue => e
          @@logger.warn "Caught #{e.message}, backtrace=#{e.backtrace}"
          raise WeaverError, e.message
        end
      end

      # Saves an infrastructure topology to the filesystem in 'filename'.  
      # If 'yaml_filename' is passed in, a yaml file of the infrastructure properties will be written, and the infrastructure will load the yaml file
      #
      # * *Args*    :
      #   - +infrastructure+ -> the topology to save (Topology)
      #   - +filename+ -> the location of the file to be saved
      #   - +yml_filename+ -> the location of the optional yml file to be saved. The location can be absolute or relative. If it is relative, it is relative to the directory of the infrastructure
      #
    def save_infrastructure(infrastructure, filename, yaml_filename = nil)
      if (filename.nil? || filename.empty? || !File.basename(filename))
        raise WeaverError.new("Filename '#{filename}' does not have a base name'")
      end

      wrote_yaml = false
      unless (yaml_filename.nil? || yaml_filename.empty?)
        if !File.basename(yaml_filename)
          raise WeaverError.new("Yaml filename '#{yaml_filename}' does not have a base name'")
        end

        relative_yaml_filename = yaml_filename
        absolute_yaml_filename = yaml_filename
        have_relative = false

        begin
          yaml_pathname = Pathname.new(yaml_filename)
          if yaml_pathname.relative?
            # if it's a relative filename, get the absolute filename
            have_relative = true
            absolute_yaml_filename = File.join(File.dirname(filename), yaml_filename)
          else
            # if it's an absolute filename, get the relative path from the infrastructure file
            yaml_path = Pathname.new(File.expand_path(File.dirname(absolute_yaml_filename)))
            inf_path = Pathname.new(File.expand_path(File.dirname(filename)))
            relative_path = yaml_path.relative_path_from(inf_path).to_s
            relative_yaml_filename = relative_path + File::SEPARATOR + File.basename(yaml_filename).to_s
            have_relative = true
          end
        rescue
          # If Pathname throws an error because a relative path cannot be calculated
          have_relative = false
        end

        # write the yaml file
        begin
          yaml_property_writer = Weaver::Utils::YamlPropertyWriter.new(infrastructure)
          yaml_property_writer.write_yaml(absolute_yaml_filename)
          wrote_yaml = true
        rescue => e
          # log the error and continue writing the infrastructure file without the yaml
          @@logger.error("Error writing yaml file #{e.message}")
          wrote_yaml = false
        end
      end
      if (wrote_yaml)
        duplicate = Weaver::Transforms::Duplicate.duplicate(infrastructure)
        yaml_property_writer.set_values_to_references(duplicate)
        File.open(filename, 'w') do |f|
#          f.write("require 'weaver/chef/chef_automation_handler'\n")
          if have_relative
#           load_string = "properties = YAML::load(File.open(File.join(File.dirname(__FILE__),'#{relative_yaml_filename}')))"
            load_string = "properties = load_properties(File.join(File.dirname(__FILE__),'#{relative_yaml_filename}'))"
          else
#           load_string = "properties = YAML::load(File.open('#{relative_yaml_filename}'))"
            load_string = "properties = load_properties('#{relative_yaml_filename}')"
          end
          f.write( load_string + "\n" + duplicate.to_weaver )
        end
      else
        File.open(filename, 'w') do |f|
          f.write( infrastructure.to_weaver )
        end
      end

    end

      private      

      # Determines the IWD version via the REST API
      # * *Returns* :
      #   - IWD version (String)
      def determine_server_variant_and_version()
        response = @client.describe_version
        if response.nil?
          @@logger.warn "Cannot determine IBM Workload Deployer version. Using '3.1' as default."
          @server_version = '3.1' # we set 3.1 as the default       
        else
          # handle version
          versionStr = response['version']
          @server_version = versionStr.split('-')[0]
                    
          # handle server variant (currently IWD or IPAS)
          case response['product']
            when 'IPAS' then 
              @server_variant = :ipas
              unless @server_version =~ /^3\.[1-9]/               
               @@logger.warn "IPAS version < 3.1 is not officially supported."
              end
            else @server_variant = :iwd
          end      
          
          # determine if cloud_group deployment is supported
          @supports_cloud_group_deployment = response['capabilities']['virtual_application_deployment_with_cloud_group']
          @@logger.debug("'Cloud group' based deployment supported: #{@supports_cloud_group_deployment}")

          # determine if environment_profile deployment is supported
          @supports_env_profile_deployment = response['capabilities']['virtual_application_deployment_with_environment_profile']
          @@logger.debug("Environment profile based deployment supported: #{@supports_env_profile_deployment}")        
        end
      end
            
      def process_node_startup_order(node, part_json_details, startup_order)
        startup_order_nr = part_json_details['startuporder']
        if startup_order[startup_order_nr].nil?
          startup_order[startup_order_nr] = node
        elsif startup_order[startup_order_nr].kind_of?(Array)
          startup_order[startup_order_nr] << node
        else
          val = startup_order[startup_order_nr]
          startup_order[startup_order_nr] = [ val, node ]
        end
      end

      # Generates the IWD server hostname with the right protocol (http || https) 
      def get_hostname() 
        if @host =~ /https/
          return @host
        else 
          return "https://#{host}"
        end        
      end
      
      # Adds ICON metadata (if available) as component to a Node
      #
      # * *Args*    :
      #   - +node+ -> the node to add the components
      #   - +virt_image_id+ -> the virtual image id in IWD
      #   - +virt_image_part_id+ -> the virtual image part id relative to the virt_image_id
      def add_icon_metadata(node, virt_image_id, virt_image_part_id)
        @@logger.debug "Processing ICON image metadata..."
        
        # primitive caching mechanism
        if icon_topologies_cached?(virt_image_id, virt_image_part_id)
          raise WeaverError.new("Aborted") if @reporter.call(Logger::INFO, "Processing", "ICCT topology information for part id: #{virt_image_part_id}") == :abort
          topologies = get_icon_topology_from_cache(virt_image_id, virt_image_part_id)
        else          
          raise WeaverError.new("Aborted") if @reporter.call(Logger::INFO, "Processing", "ICCT topology information for part id: #{virt_image_part_id}") == :abort
          topologies = @client.describe_virtual_image_topologies(virt_image_id, virt_image_part_id) # returns [] if no topologies available.
          cache_icon_topology(virt_image_id, virt_image_part_id, topologies)
        end
        
        automation_xml_doc = nil, semantic_xml_doc = nil
        topologies.each do |top|
          if top['topType'] == 'automation'
            automation_xml_doc = REXML::Document.new(top['topXML'])
            #@@logger.debug "automation topology: #{automation_xml_doc}"
          elsif top['topType'] == 'semantic'            
            semantic_xml_doc = REXML::Document.new(top['topXML'])
            #@@logger.debug "semantic topology: #{semantic_xml_doc}"   
          end
        end 
        
        if automation_xml_doc && semantic_xml_doc 
          
          generator = Weaver::Zephyr::WeaverGenerator.new(semantic_xml_doc, :cloud_pattern, {})
          generator.transform_document(automation_xml_doc)
          
          if generator.root.env
            inf = generator.root.env.infrastructure
          elsif generator.root.infrastructure
            inf = generator.root.infrastructure
          end
            
          unless inf.nil?
            Weaver::Core::WeaverTransforms.improve_ids(inf) 
            
            inf.nodes.each do |n_id, n|
              n.components.each do |c_id, c|
                dup = c.clone
                dup.scope = node
                node.components[c_id] = dup 
              end
            end
          end  
      
        end             
      end
      
      def icon_topologies_cached?(virt_image_id, virt_image_part_id)
        id = "#{virt_image_id}_#{virt_image_part_id}"
        cached = @icon_topology_cache != nil && @icon_topology_cache[id] != nil
        @@logger.debug "Checking if ICON image topology '#{id}' is cached...#{cached}"
        return cached
      end
      
      def get_icon_topology_from_cache(virt_image_id, virt_image_part_id)
        id = "#{virt_image_id}_#{virt_image_part_id}"
        @@logger.debug "Retrieving ICON image topology '#{id}' from cache..."
        return @icon_topology_cache[id]
      end
      
      def flush_icon_topologies_cache()
        @icon_topology_cache = {}
      end

      def cache_icon_topology(virt_image_id, virt_image_part_id, topologies)
        id = "#{virt_image_id}_#{virt_image_part_id}"
        @@logger.debug "Caching ICON image topology '#{id}'..."
        @icon_topology_cache[id] = topologies
      end

      
      # Defines a cloud group attribute and environment_profile on the infrastructure and adds all the options as a comment
      #
      # * *Args*    :
      #   - +inf+ -> the current infrastructure instance
      #
      def add_cloud_group_and_env_profile_attr(inf)        
        
        # cloud group attr if supported
        if @supports_cloud_group_deployment
          cloud_attr = create_cloud_group_attr(inf, @supports_env_profile_deployment)
          cloud_attr.prefix_comments.insert 0, "\n# Select a cloud group or an environment profile.\n#"
          inf.amd[:cloud_group] = cloud_attr
          inf.amd[:cloud_group].add_validator(:exactly_one_value, :environment_profile) if @supports_env_profile_deployment # add validator only if environment_profile is supported     
        end 
        
        # environment_profile attribute
        if @supports_env_profile_deployment
          begin 
            env_profiles = @client.describe_environment_profiles

            env_profile_attr = Attribute.new(:environment_profile, inf, true)
            env_profile_attr.name = "environmentProfile"
            env_profile_attr.implicit = false
            env_profile_attr.required = false if @supports_cloud_group_deployment # set required if cloud_group is not present
            inf.amd[:environment_profile] = env_profile_attr
    
            ids = []
            raise WeaverError.new("Aborted") if @reporter.call(Logger::DEBUG, "Adding", "Environment profiles") == :abort
            env_profile_attr.prefix_comments << "\n# Available environment profiles:"
            env_profiles.each do |p|
              ids << p['id']
              env_profile_attr.prefix_comments << "id #{p['id']}: #{p['name']}"
            end
            ids << nil if @supports_cloud_group_deployment
            # set first one as default if we did not already set cloud_group as default
            env_profile_attr.value = ids[0] if ids.size >= 1 && !@supports_cloud_group_deployment 
            inf.amd[:environment_profile].add_validator(:valid_values, *ids)
            inf.amd[:environment_profile].add_validator(:exactly_one_value, :cloud_group) if @supports_cloud_group_deployment
          rescue DevopsDeploymentLib::HttpError => e
            @@logger.warn "User is not allowed to read environment profile information. Skipping..."
            @env_profile_readable = false
          end            
        end
      end
      
      def add_cloud_group_and_ip_group(node)
        unless @supports_env_profile_deployment && @env_profile_readable
          return # we don't need the attributes if we environment profiles are not supported or user is not allowd to read them
        end
        
        raise WeaverError.new("Aborted") if @reporter.call(Logger::DEBUG, "Adding",  "Cloud and IP groups") == :abort
        
        begin 
          cloud_attr = create_cloud_group_attr(node, @supports_cloud_group_deployment)
          cloud_attr.prefix_comments.insert 0, "\n# Set the cloud_group and ip_group value if you are deploying using an environment profile.\n#"
          cloud_attr.add_validator(:not_empty_if, :environment_profile)
          cloud_attr.required = false if @supports_cloud_group_deployment
          node.amd[:cloud_group] = cloud_attr
          cloud_attr.value = nil if @supports_cloud_group_deployment
          
          # add ip_groups
          ip_group_attr = Attribute.new(:ip_group, node, true)
          ip_group_attr.name = "ipGroup"
          ip_group_attr.implicit = false
          ip_group_attr.required = false if @supports_cloud_group_deployment
          node.amd[:ip_group] = ip_group_attr
            
          ids = []
          ip_groups = @client.describe_ip_groups
          ip_group_attr.prefix_comments << "\n# Available IP Groups:"
          ip_groups.each do |p|
            ids << p['id']
            ip_group_attr.prefix_comments << "id #{p['id']}: #{p['name']}"
          end
          ids << nil if @supports_cloud_group_deployment # nil is a valid if only if cloud_group and environment profiles are possible                  
          ip_group_attr.value = ids[0] if ids.size >= 1 and !@supports_cloud_group_deployment
         
          node.amd[:ip_group].add_validator(:valid_values, *ids)
          node.amd[:ip_group].add_validator(:not_empty_if, :environment_profile)
          # add extra_ip_groups
          extra_ip_groups_attr = Attribute.new(:extra_ip_groups, node, true)
          extra_ip_groups_attr.name = "ipGroup"
          extra_ip_groups_attr.implicit = false
          extra_ip_groups_attr.required = false if @supports_cloud_group_deployment
          node.amd[:extra_ip_groups] = extra_ip_groups_attr 
        rescue DevopsDeploymentLib::HttpError => e
          @@logger.warn "User is not allowed to read environment profile information. Skipping..."
        end   
      end

      def create_cloud_group_attr(parent, nil_allowed = true)
        raise WeaverError.new("Aborted") if @reporter.call(Logger::DEBUG, "Inspecting", "Cloud groups") == :abort
        cloud_attr = Attribute.new(:cloud_group, parent, true)
        cloud_attr.implicit = false
        cloud_attr.name = "cloud"
        ids = []
        cloud_groups = @client.describe_cloud_groups
        cloud_attr.prefix_comments << "Available cloud groups:"
        cloud_groups.each do |group|
          ids << group['id']
          cloud_attr.prefix_comments << "id #{group['id']}: #{group['name']}"
        end
        ids << nil if nil_allowed
        cloud_attr.add_validator(:valid_values, *ids)
        cloud_attr.value = ids[0] if ids.size >= 1  # set first one as default
        return cloud_attr
      end         
   
      # Adds a script package to a Weaver iwd_node.
      # * *Args*    :
      #   - +node+ -> the iwd_node for adding the script package 
      #   - +pattern_id+ -> the id of the pattern 
      #   - +part_id+ -> the id of the pattern part
      def add_script_packages(node, pattern_id, part_id)
        raise WeaverError.new("Aborted") if @reporter.call(Logger::DEBUG, "Adding", "Script packages") == :abort
        scripts = @client.describe_part_scripts(pattern_id, part_id)
        
        script_order = [] # script ordering
        unless scripts.nil?
          scripts.each do |script_json|
            script_pkg = @root.prototype_instance(:script_package, script_json['key'].symbolize, node, nil)
            script_pkg.name = script_json['label']
            script_pkg.description = script_json['description']
            script_pkg.script_id= script_json['scriptId'].to_i
            script_pkg.id_in_pattern= script_json['id'].to_i
              
           
            script_def_json = @client.describe_script_package(script_json['scriptId'])             
            script_pkg.archive = script_def_json['filename'] if script_def_json
            #puts script_def_json

            # add script package attributes as attributes to component
            script_json['parameters'].each do |para_json|
              #puts para_json.to_json
              attr = add_attributes(script_pkg, para_json)
              script_pkg.amd[attr.id] = attr unless attr.nil?
            end
            
            order_number = script_json['executionOrder']
            #puts "order number #{order_number}"
            if order_number == 0 || !script_order[order_number].nil?
              script_order << script_pkg
            else
              script_order[order_number] = script_pkg
            end
            
            # Add script package to node   
            node.automation[script_pkg.id] = script_pkg
          end
        end
        # set script order
        node.automation_order= script_order.select { |ele| ele }
        
      end

      # Adds an attribute/property to any given model element. An attribute is only added if it
      # is set as 'userConfigurable' in IWD.
      # * *Args*    :
      #   - +model_element+ -> the Weaver model element the attribute needs to be attached to
      #   - +prop_json+ -> the JSON element containing the property
      def add_attributes(model_element, prop_json)
        user_configurable = prop_json['userConfigurable']
        if user_configurable.class == String # Workaround because the value is supposed to be a boolean but is often a string
          user_configurable = user_configurable.to_b
        end
        if user_configurable && prop_json['requiredForDeployment']
          attr = Attribute.new(prop_json['key'].symbolize, model_element)
          attr.name = prop_json['key']
          attr.label = prop_json['label'] if prop_json['label'] != prop_json['description']
          attr.description = prop_json['description']
          attr.required = prop_json['requiredForDeployment']
          attr.mutable = true
          attr.implicit = false
          if prop_json['password'] == true || prop_json['password'] == 'true' # the boolean values in this JSON are often strings
            is_password = true
            attr.tags << :password 
          end
          value = prop_json['value'] || prop_json['defaultvalue']
          
          type = prop_json['type']
          if type == 'boolean'
            value = value.to_b
          elsif type == 'integer'
            value = value.to_i
          end 
          
          unless value.nil? || (value.kind_of?(String) && value.empty?)
            attr.default = value
          end
          
          attr.group = prop_json['pclass']

          values = []
          prop_json_arr = prop_json['validValues'];
          unless prop_json_arr.nil?                      
            prop_json_arr.each do |v| 
              if value.kind_of?(Fixnum)
                values << v.to_i
              else
                values << v
              end            
            end 
          end
          
          # valid_values only if multiSelect is not enabled
          multiSelect = prop_json['multiSelect'] || false
          attr.add_validator(:valid_values, *values) if !values.empty?() && !multiSelect
          
          # this is necessary if we have more than one value that can be selected
          attr.add_validator(:subset, *values) if !values.empty?() && multiSelect
          
          # add validators for value types
          attr.add_validator(:value_type, :numeric) if value.kind_of? Fixnum
          attr.add_validator(:value_type, :boolean) if value == true || value == false
          attr.add_validator(:value_type, :string) if value.kind_of? String
          attr.add_validator(:value_type, :password) if is_password
          attr.add_validator(:value_type, :array) if value.kind_of? Array
          
          # hack to avoid duplicate keys by appending _{i}
          i = 0
          while model_element.amd[attr.id]
            attr.id = "#{attr.id}_#{i}".symbolize
            i += 1
          end
          model_element.amd[attr.id] = attr
          return attr
        end
        return nil
      end

      def add_os_information(node, virtual_image_id)
        raise WeaverError.new("Aborted") if @reporter.call(Logger::DEBUG, "Retrieving", "OS information") == :abort
        vi_json = @client.describe_virtual_image(virtual_image_id)
        #@@logger.debug("--start virtual image------------")
        #@@logger.debug(vi_json.to_json)
        #@@logger.debug("--end virtual image------------")

        os_handler = Weaver::Core::Os::OperatingSystemHandler.new
        os_info = os_handler.match_os(vi_json['operatingsystemdescription'])
        os = OperatingSystem.new()
        os.type= os_info[:type] unless os_info.nil?
        os.flavor= os_info[:flavor] unless os_info.nil?
        os.version= vi_json['operatingsystemversion']
        os.architecture=os_handler.match_architecture(vi_json['operatingsystemid_text'])
        node.os = os
      end

    end

  end
end
