require 'rest_client'
require 'weaver/service/rest_client_base'
require 'json'
require 'weaver/core/json'
require 'weaver/environment'
require 'weaver/utils/environment_util'
require 'weaver/service/weaver_service_error'
require 'hash'
require 'weaver/deployers/execution/continuous_validation_setup'

module Weaver
  module Service
    # Implements methods that make REST calls to the RestEndpoint of the Weaver Service.
    # The class defines methods that makes REST calls, and interpret the REST responses into simpler objects
    class ServiceClient < Weaver::Service::RestClientBase
      include Weaver::Core
      
      def initialize(host, port)
        host = host
        port = port
        super host,port
      end

      # Post an environment to the database
      # * *Args*    :
      #   - +root_or_file+ -> the Weaver model of an environment root or the file location of the environment. If
      #       a filename is passed, the name of the file will be store in the environment resource
      #   - +model_state+ -> the state of the model, "predeployment" or" deployed"
      # * *Returns*
      #  The UUID of the environment in CouchDB
      # * *Raises*
      # WeaverServiceError if there is a problem
      def post_environment(root_or_file, model_state=nil)
        begin
          if root_or_file.kind_of?(Root)
            #TODO: Hardcoding the Weaver project directory for now. It will be passed as a parameter to the post /deployments call.
            continuous_val_setup = Weaver::Deployers::Execution::ContinuousValidationSetup.new(root_or_file, "/root/devops/experimental/com.ibm.devops.weaver.samples/rubis-openstack-contval-test")
            continuous_val_setup.add_rules_from_files()
            root_json = JSON.generate(root_or_file)
            params = {:environment_root => root_json}
          else
            environment_file = root_or_file
            unless File.exists?(environment_file)
              raise WeaverServiceError, "Environment file does not exist: #{environment_file}"
            end
            params = {:source_filename => environment_file}
          end
          params[:model_state] = model_state if (!model_state.nil?)
          url = "/environments"
          response = post(url, params)
          if (response.code == 201)
            json = JSON.parse(response.body)
            return json["id"]
          else
            raise WeaverServiceError.new("Error posting environment, response code is #{response.code}", response.code)
          end
        rescue WeaverServiceError => e
          raise e
        rescue => e
          raise  WeaverServiceError.new("Failed to post environment: #{e.inspect}", nil, e)
        end
      end

      # Get information about an environment that is stored in the database
      # * *Args*    :
      #   - +environment_id+ -> the id of the environment to get (String)
      # * *Returns*
      #  A Hash of environment information:
      #    "environment_root" => a Hash representation of the root of the environment model
      #    "source_filename" [optional] => the name of the Weaver file of the environment.
      #       May be nil if the environment resource was created from a root instead of from a filename
      #    "display_name" => the display name of the environment used to identify the environment to the user
      #    "creation_date" => the date the environment document was created in the database
      #    "model_state" [optional] => the state of the model, "predeployment" or "deployed"
      #
      # TODO: return an EnvironmentInfo object that has accessor methods instead of returning a hash
      def get_environment_info(environment_id)
        begin
          url = "/environments/#{environment_id}"
          response = get(url)
          json_hash = JSON.parse(response.body)
          if (response.code == 200)
            return json_hash;
          else
            raise WeaverServiceError.new("Failed to get environment #{environment_id}, response code is #{response.code}", response.code)
          end
        rescue WeaverServiceError => e
          raise e
        rescue => e
          raise  WeaverServiceError.new("Failed to get environment #{environment_id}: #{e.inspect}", nil, e)
        end

      end

      # Get an environment from the database
      # * *Args*    :
      #   - +environment_id+ -> the id of the environment to get (String)
      # * *Returns*
      #  The root Weaver model of the environment
      def get_environment_root(environment_id)
        begin
          params = get_environment_info(environment_id)
          root_hash = params["environment_root"]
          root = Root.from_hash(root_hash)
          return root
        rescue WeaverServiceError => e
          raise e
        rescue => e
          raise  WeaverServiceError.new("Failed to get environment #{environment_id}: #{e.inspect}", nil, e)
        end

      end

      # Post a deployment to the database
      # * *Args*    :
      #   - +root_or_file+ -> the Weaver model of an environment root or the file location of the environment. If
      #       a filename is passed, the name of the file will be store in the environment resource
      #   - +properties+ -> a Hash of properties that are needed to configure the underlying deployer.
      #                     This is implementation specific and may contain username, password, key files, etc.
      #   - +deploy_parameters+ -> a Hash deployment time parameters for the environment topology (e.g., :deps_file_url)
      #   - +environment_variables+ -> a Hash of environment variables/values to set in the environment before deployment
      #   - +validate+  -> if true, will run validation before posting the environment to be deployed. Although validation
      #        will run right before deployment, the user may want to run validation here to prevent the environemnt from
      #        being posted in the case of a validation that will fail.
      # * *Returns*
      #  The UUID of the deployment in CouchDB
      def post_deployment(root_or_file, properties, deployment_parameters, environment_variables, validate = false)
        properties = {} if properties.nil?
        deployment_parameters = {} if deployment_parameters.nil?
        environment_variables = {} if environment_variables.nil?
        original_deployment_parameters = deployment_parameters.clone # validation modifies the deployment_parameters hash, so the original set of params is saved here

        begin
          if (validate)
            run_validation(root_or_file, properties, deployment_parameters, environment_variables)
          end

          environment_id = post_environment(root_or_file, "predeployment")

          url = "/deployments"
          params = {:environment_id => environment_id, :properties => properties.to_json, :deployment_parameters => original_deployment_parameters.to_json, :environment_variables => environment_variables.to_json}
          response = post(url, params)
          if (response.code == 201)
            json = JSON.parse(response.body)
            return(json["id"] )
          else
            raise WeaverServiceError.new("Failed to post deployment, response code is #{response.code}", response.code)
          end
        rescue WeaverServiceError => e
          raise e
        rescue => e
          raise  WeaverServiceError.new("Failed to post deployment: #{e.inspect}", nil, e)
        end
      end

      def update_deployment(deployment_id, deployed_environment_id)
        params = {}
        params["deployed_environment_id"] = deployed_environment_id
        begin
          url = "/deployments/#{deployment_id}"
          params = {"deployed_environment_id" => deployed_environment_id}
          response = put(url, params)
          if (response.code == 201)
            json = JSON.parse(response.body)
            return(json["id"] )
          else
            raise WeaverServiceError.new("Error updating deployment, response code is #{response.code}", response.code)
          end
        rescue WeaverServiceError => e
          raise e
        rescue => e
          raise  WeaverServiceError.new("Failed to update deployment: #{e.inspect}", nil, e)
        end
      end

      # Get an environment from the database
      # * *Args*    :
      #   - +deployment_id+ -> the id of the deployment to get (String)
      # * *Returns*
      #  A Hash of deployment information:
      #   "environment_id" => the database id of the environment to deploy
      #   "properties" => the properties to pass to the deployer
      #   "deployment_parameters" => the deployment parameters to pass to the deployer
      #        (note: keys will be converted from the json string representation to symbols, as expected by the deployer)
      #   "environment_variables" => environment variables to set before deployment
      #        (note: keys will be converted from the json string representation to symbols, as expected by the deployer)
      #
      def get_deployment(deployment_id)
        begin
          url = "/deployments/#{deployment_id}"
          response = get(url)
          json_hash = JSON.parse(response.body)
          if (response.code == 200)
            deployment_hash = {}
            symbolize = ["properties", "deployment_parameters"]
            json_hash.each do |key, value|
              if symbolize.include?(key)
                deployment_hash[key] = json_hash[key].symbolize
              else
                deployment_hash[key] = json_hash[key]
              end
            end
            return deployment_hash
          else
            raise WeaverServiceError.new("Failed to get deployment #{deployment_id}, response code is #{response.code}", response.code)
          end
        rescue WeaverServiceError => e
          raise e
        rescue => e
          raise  WeaverServiceError.new("Failed to get deployment #{deployment_id}: #{e.inspect}", nil, e)
        end

      end

      # Run validation on an environment. Environment variables will be set before validation is run.
      # * *Args*    :
      #   - +root_or_file+ -> the Weaver model of an environment root or the file location of the environment
      #   - +properties+ -> a Hash of properties that are needed to configure the underlying deployer.
      #                     This is implementation specific and may contain username, password, key files, etc.
      #   - +deploy_parameters+ -> a Hash deployment time parameters for the environment topology (e.g., :deps_file_url)
      #   - +environment_variables+ -> a Hash of environment variables/values to set in the environment before deployment
      # * *Raises*
      # ValidationError if validation fails
      def run_validation(root_or_file, properties, deployment_parameters, environment_variables)
        root = get_root(root_or_file)
        env = Weaver::Environment.new(root)
        hold_variables = Weaver::Utils::EnvironmentUtil.set_environment_variables(environment_variables)
        env.validate([], properties, deployment_parameters)
        Weaver::Utils::EnvironmentUtil.set_environment_variables(hold_variables)
      end

      # Returns the root, given a root or a filename. If it's a filename, the Root will be parsed from the file
      # * *Args*    :
      #   - +root_or_file+ -> the Weaver model of an environment root or the file location of the environment
      # * *Returns*
      #  The Weaver model of the environment rot
      def get_root(root_or_file)
        if root_or_file.kind_of?(Root)
          return root_or_file;
        else
          environment_file = root_or_file
          unless File.exists?(environment_file)
            raise Weaver::Core::WeaverServiceError, "Environment file does not exist: #{environment_file}"
          end
        end
      end

    end
  end

end