require 'singleton'
require 'weaver/core/logging'

module Weaver
    # Implements a singleton for Weaver global properties.
    #
    # This class treats property keys as symbols and strings interchangeably.
    #
    # For instance, one can do set_property(:my_prop, "my_value") and later get_property("my_prop") returns "my_value"
    class System
      include Singleton
      include Weaver::Core

      WEAVER_VERSION_FILE = File.expand_path(File.join(File.dirname(__FILE__), '..', '..', 'VERSION'))
            
      # Keeps the default Weaver properties. New defaults must be added here.
      DEFAULT_PROPS = { }
      
      def initialize
        @weaver_properties = Hash.new
        
        ver = read_weaver_version
        unless ver.nil?
          DEFAULT_PROPS["version"] = ver 
        end
        
        DEFAULT_PROPS.each do | name, value |
          @@logger.debug "Setting default WeaverProperty '#{name}' to '#{value}'"
          set_property(name, value)
        end
        
      end
      
      # Returns the value of the property with the given name, or nil if an illegal argument is passed.
      # The argument prop_name must be either a Symbol or a String
      def get_property(prop_name)
        return nil if (!prop_name.kind_of?(String) && !prop_name.kind_of?(Symbol))
        name = prop_name.kind_of?(Symbol) ? prop_name.to_s : prop_name
        @weaver_properties[name]
      end
      
      # Returns a hash string => string containing all (key,value) pairs
      #
      def get_properties_as_string_hash()
        @weaver_properties 
      end
      
      # Sets the value of a property
      # The argument prop_name must be either a Symbol or a String
      # The argument prop_value must be a string
      # This method returns the value assigned to the property, or nil if ilegal arguments are passed.
      def set_property(prop_name, prop_value)
        return nil if (!prop_name.kind_of?(String) && !prop_name.kind_of?(Symbol))
        # Commented out by FR: too strict
        #return nil if !prop_value.kind_of?(String) 
        name = prop_name.kind_of?(Symbol) ? prop_name.to_s : prop_name
        @weaver_properties[name] = prop_value
        prop_value
      end
      
      private
      
      def read_weaver_version
        if File.exist?(WEAVER_VERSION_FILE)
          version = nil
          File.open(WEAVER_VERSION_FILE, 'r') do |f|
            version = f.read
          end 
          return version.strip
        else 
          @@logger.error "Could not find the Weaver VERSION file '#{WEAVER_VERSION_FILE}"
        end
      end
    end
end