require 'open3'
require 'socket'
require 'weaver/continuous_validation'
require 'weaver/continuous_validation/rule_logger'
require 'weaver/continuous_validation/service_client'

module Weaver
  module ContinuousValidation
    class RuleRunner
      def initialize(deployment_id, rule_id, node_id, command, service_host, service_port)
        @deployment_id = deployment_id
        @rule_id = rule_id
        @node_id = node_id
        @command = command
        
        @rule_log_dir = File.join(LOG_DIRECTORY, @rule_id)
        FileUtils.mkdir_p(@rule_log_dir)
        @rule_stdout_filename = File.join(@rule_log_dir, @rule_id + '.out')
        @rule_stderr_filename = File.join(@rule_log_dir, @rule_id + '.err')
        rule_log_filename = File.join(@rule_log_dir, @rule_id + '.log')
        @rule_logger = RuleLogger.new(rule_log_filename, VALIDATION_HISTORY_SIZE)
        
        @service_client = ServiceClient.new(service_host, service_port)
      end 
      
      def run()
        # Run rule's command 
        $logger.debug "Wrapper running command '#{@command}' for rule '#{@rule_id}'"
        begin
          Open3.popen3(@command) do |stdin, stdout, stderr, process|
            exit_code = process.value

            # Captures command's stdout
            stdout_lines = stdout.readlines
            f = File.open(@rule_stdout_filename, 'w')
            f.write(stdout_lines.join) unless stdout_lines.size == 0
            f.close()

            # Captures command's stderr
            stderr_lines = stderr.readlines
            f = File.open(@rule_stderr_filename, 'w') 
            f.write(stderr_lines.join) unless stderr_lines.size == 0
            f.close()

            success, logger_message, service_message = process_validation_result(stdout_lines, stderr_lines, exit_code)
            report_error_if_needed(success, logger_message, service_message)
          end
        rescue Errno::ENOENT => e
          # The rule's command could not be executed by the OS because it either does not exist or could not be found.
          $logger.error "Error while running command '#{@command}' for rule '#{@rule_id}': '#{e}'"
          
          # Truncate stdout and stderr files to avoid confusion with a possibly successful previous run.
          f = File.open(@rule_stdout_filename, 'w')
          f.close()
          f = File.open(@rule_stderr_filename, 'w')
          f.close()
          
          #TODO Report validation error to the weaver service
        end
      end
      
      private
      
      def process_validation_result(stdout_lines, stderr_lines, exit_code)
        # Try to match the last line of the command's stdout against VALIDATION_RESULT_FORMAT
        match = get_validation_result(stdout_lines)
        
        # Try to match the last line of the command's stderr, if needed
        match = get_validation_result(stderr_lines) if match.nil?
        
        success = true
        logger_message = service_message = SUCCESS_STR
        if match.nil?
          # We need to rely on the command's exit code
          if exit_code.exitstatus != 0
            # Validation error
            success = false
            service_message = "Exit code = #{exit_code.exitstatus}"
            logger_message = generate_error_message(service_message)
            $logger.debug "Command '#{@command}' for rule '#{@rule_id}' resulted in validation error due to exit code. Message: #{logger_message}"            
          end
        else
          # We can rely on the last line of the command's stdout or stderr
          if match[1] == 'ERROR'
            # Validation error
            success = false
            service_message = match[2]
            logger_message = generate_error_message(match[2])
            $logger.debug "Command '#{@command}' for rule '#{@rule_id}' resulted in validation error due to output match. Message: #{logger_message}"
          end
        end
        
        return success, logger_message, service_message
      end
      
      def generate_error_message(text = nil)
        msg = "ERROR in deployment #{@deployment_id}. Rule '#{@rule_id}' failed on node whose IPs are '#{get_ip_info.map{|info| info.ip_address}.inspect}'."
        msg += " Error message: '#{text}'" unless text.nil? || text == ""
        return msg
      end
      
      def get_ip_info()
        addr_info = Socket.ip_address_list.select{|i| i.ipv4? && !i.ipv4_loopback? && !i.ipv4_multicast?}
        return addr_info || []
      end
           
      def get_validation_result(stdout_or_stderr_lines)
        size = stdout_or_stderr_lines.size
        if size > 0
          last_line = stdout_or_stderr_lines[size - 1]
          match = last_line.match(VALIDATION_RESULT_FORMAT)
        else
          match = nil
        end   
        
        return match
      end

      def report_error_if_needed(success, logger_message, service_message)
        last_result = @rule_logger.get_last_validation_result
        begin
          if logger_message != last_result
            @rule_logger.log_validation_result(logger_message)
            if success == false
              # The rule is failing; it was possibly succeeding before.
              # Let's tolerate up to VALIDATION_ERROR_THRESHOLD failures before triggering an alert.
              error_count = get_and_update_error_count()
              if error_count > VALIDATION_ERROR_THRESHOLD && error_count != (VALIDATION_ERROR_THRESHOLD + 2)
                $logger.debug "About to post a validation result for rule '#{@rule_id}'."
                @service_client.post_validation_result(@deployment_id, @rule_id, @node_id, service_message)
                $logger.debug "Posted validation result for rule '#{@rule_id}'."
              end
            end
            #TODO: Handle the case where a previously failing rule now succeeds, i.e., success == true
            #      We may need to delete a previously posted validation error.
          else
            if success == false
              # The rule is failing.
              # Let's tolerate up to VALIDATION_ERROR_THRESHOLD failures before triggering an alert.
              error_count = get_and_update_error_count() 
              if error_count > VALIDATION_ERROR_THRESHOLD && error_count != (VALIDATION_ERROR_THRESHOLD + 2)
                $logger.debug "About to post a validation result for rule '#{@rule_id}'."
                @service_client.post_validation_result(@deployment_id, @rule_id, @node_id, service_message)
                $logger.debug "Posted validation result for rule '#{@rule_id}'."
              end
            end
          end
        rescue WeaverServiceError => e
          $logger.error("Error making a REST call against the #{SERVICE_VALIDATION_URI} resource of the Weaver service: message is #{e}, response code is #{e.response_code}, exception class is #{e.class}, exception is #{e.inspect}")
        end
      end
      
      def get_and_update_error_count()
        # Right now we keep one error count file for the entire rule.
        # We may want to keep multiple error count files, one per error message produced by the same rule.
        #In this case, the filenames should be generated based on a hash function applied to the error message.
        filename = File.join(@rule_log_dir, ERROR_COUNT_LOG_PREFIX + ".log")
        retval = 0
        if File.exist?(filename)
          f = File.open(filename, 'r')
          retval = f.gets.chomp.to_i
          f.close()
        end
        retval += 1
        if retval <= VALIDATION_ERROR_THRESHOLD + 1
          f = File.open(filename, 'w')
          f.puts(retval)
          f.close()
        end
        return retval
      end
      
    end
  end
end