require 'logger'
require 'weaver/coordination/logging'
require 'weaver/coordination/coordinated_transaction_manager'

module Weaver
  module Coordination
      
    # Create a kind of late binding to be evaluated at runtime on a VM.
    class VmLateBinding
      include Weaver::Coordination::Logging

      def initialize(expression, values, attr_mapping)
#        @@logger.debug "creating proc with #{expression}"
        @proc = Kernel.eval("#{expression}")
        @values = values
        @attr_map = attr_mapping
        @extra_args = []
      end
      
      # Evaluate the late binding in a sandbox
      # *Args*
      #  - +transaction+ the coordination transaction
      def eval(transaction)
        sandbox = VmSandbox.new(@values, @attr_map, transaction, '', [])
        retval = sandbox.instance_exec(@extra_args, &@proc)
        return retval
      end # def eval
    end # class VmLateBinding

    # This class implements a 'sandbox' for in which we evaluate late_binding expressions 
    # that need to be evaluated on a target VM.
    class VmSandbox
      include Weaver::Coordination::Logging

      # *Args*
      #  - +all+
      #  - +values+ a Hash mapping fully_qualified_ids of of weaver Base objects to values
      #  - +transaction+ transaction to use when requesting values from the coordinator
      #  - +obj_ref+ a fully qualified id to a weaver object 
      #  - +all_values+ cache of values for evaluation of all()
      def initialize (values, attr_mapping, transaction, obj_ref = '', all_values = [])
        @values = values
        @attr_map = attr_mapping
        @transaction = transaction
        @obj_ref = obj_ref || ''
        @all_values = all_values || []
#          
#        @@logger.debug "created sandbox with @values = #{@values}"
#        @@logger.debug "created sandbox with @attr_map = #{@attr_map}"
#        @@logger.debug "created sandbox with @coordination_id = #{@coordination_id}"
#        @@logger.debug "created sandbox with @obj_ref = #{@obj_ref}"
#        @@logger.debug "created sandbox with @all_values = #{@all_values}"
      end
      
      # Handle all() called with an attribute parameter. Instead of returning the current value (arg), 
      # lookup precomputed value.
      def all(arg)
#        @@logger.debug "all called; returning #{@all_values}"
        return @all_values.clone
      end # def all
      
      def method_missing(sym, *args, &block)
#        @@logger.debug "method_missing called with #{sym.inspect}"
        
        new_obj_ref = "#{@obj_ref}"
        new_obj_ref << '.' unless @obj_ref.empty?
        new_obj_ref << sym.to_s

        # Check if the obj_ref is something we recognize as an entry in our mapping
        lookup_key = new_obj_ref        
        if @attr_map.key?(lookup_key)
#          @@logger.debug "Lookup up value; key is: #{lookup_key} --> #{@attr_map[lookup_key]}"
          fqids = @attr_map[lookup_key]
          obj_values = fqids.map do |ref|
            obj_value = @values[ref]
#            @@logger.debug "   found value #{obj_value}"
            if obj_value.nil?
              # value was not available when called from weaver; read from coordinator
#              @@logger.debug  "lookup #{ref} via coordinator (#{@transaction})"
              obj_value = @transaction.read_data(ref)
            end # if obj_value.nil?
#            @@logger.debug "   final value is value #{obj_value}"
            obj_value
          end # fqids.map do
          @all_values.clear
          @all_values.concat(obj_values)
          # return just 1 value; remainder returned if all() called
          # @@logger.debug "    returning first element of #{@all_values.first}"
          return @all_values.first 
        end # if @attr_map.key?(lookup_key)
        
        # It wasn't something we could look up, continue to the next symbol
        # go on with the evaluation
        return VmSandbox.new(@values, @attr_map, @transaction, new_obj_ref, @all_values)
      end # def method_missing
      

    end # def VmSandbox
      
  end # module Coordination
end # module Weaver